# standard library
import unittest
import io
import sys

# third party imports
import numpy as np

# local imports
from probeye.definition.inference_problem import InferenceProblem
from probeye.definition.forward_model import ForwardModelBase
from probeye.definition.sensor import Sensor
from probeye.definition.noise_model import NoiseModelBase


class TestProblem(unittest.TestCase):

    def test_properties(self):
        p = InferenceProblem("TestProblem")
        # add first latent parameter; n_latent_prms should be 1
        p.add_parameter('a', 'model',
                        prior=('normal', {'loc': 0.0, 'scale': 1.0}))
        self.assertEqual(p.n_latent_prms, 1)
        # add second latent parameter; n_latent_prms should be 2
        p.add_parameter('b', 'noise',
                        prior=('normal', {'loc': 0.0, 'scale': 1.0}))
        self.assertEqual(p.n_latent_prms, 2)
        # check the different properties
        self.assertEqual(p.n_prms, 6)
        self.assertEqual(
            set(p.prms), {'loc_a', 'scale_a', 'a', 'loc_b', 'scale_b', 'b'})
        self.assertEqual(p.n_latent_prms, 2)
        self.assertEqual(set(p.latent_prms), {'a', 'b'})
        self.assertEqual(p.n_constant_prms, 4)
        self.assertEqual(set(p.constant_prms),
                         {'loc_a', 'scale_a', 'loc_b', 'scale_b'})
        self.assertEqual(p.n_model_prms, 1)
        self.assertEqual(set(p.model_prms), {'a'})
        self.assertEqual(p.n_prior_prms, 4)
        self.assertEqual(set(p.prior_prms),
                         {'loc_a', 'scale_a', 'loc_b', 'scale_b'})
        self.assertEqual(p.n_noise_prms, 1)
        self.assertEqual(set(p.noise_prms), {'b'})

    def test_info(self):
        # simply check that no errors occur when info-method is called; the
        # output that is usually printed is redirected
        p = InferenceProblem("TestProblem")
        p.add_parameter('a', 'model', const=1.0)
        p.add_parameter('b', 'model', prior=('normal', {'loc': 0, 'scale': 1}))
        p.add_parameter('s', 'noise', prior=('normal', {'loc': 0, 'scale': 1}))
        sys.stdout = io.StringIO()
        # try out different options
        p.info(include_experiments=False, tablefmt="presto",
               check_consistency=False)
        _ = p.info(include_experiments=False, tablefmt="presto",
                   check_consistency=False)
        p.info(include_experiments=True, tablefmt="presto",
               check_consistency=False)
        p.info(include_experiments=True, tablefmt="plain",
               check_consistency=False)
        sys.stdout = sys.__stdout__  # reset printout to console
        with self.assertRaises(AssertionError):
            # the problem is not consistent yet (e.g. no forward model defined
            # yet), so the consistency_check will raise an error
            p.info(include_experiments=True, tablefmt="presto",
                   check_consistency=True)
        # now add the remaining stuff to make to problem consistent
        test_model = ForwardModelBase('b', Sensor('x'), Sensor('y'))
        p.add_forward_model('TestModel', test_model)
        p.add_noise_model(NoiseModelBase('normal', 's', sensors=Sensor('y')))
        p.add_experiment('Experiment_1', sensor_values={'x': 1, 'y': 1},
                         fwd_model_name='TestModel')
        sys.stdout = io.StringIO()
        # now, the consistency_check should not raise an error
        p.info(include_experiments=True, tablefmt="presto",
               check_consistency=True)
        sys.stdout = sys.__stdout__  # reset printout to console

    def test_str(self):
        # set up a consistent problem and print it
        p = InferenceProblem("TestProblem")
        p.add_parameter('a', 'model', const=1.0)
        p.add_parameter('b', 'model', prior=('normal', {'loc': 0, 'scale': 1}))
        p.add_parameter('s', 'noise', prior=('normal', {'loc': 0, 'scale': 1}))
        test_model = ForwardModelBase('b', Sensor('x'), Sensor('y'))
        p.add_forward_model('TestModel', test_model)
        p.add_noise_model(NoiseModelBase('normal', 's', sensors=Sensor('y')))
        p.add_experiment('Experiment_1', sensor_values={'x': 1, 'y': 1},
                         fwd_model_name='TestModel')
        sys.stdout = io.StringIO()  # redirect output to console
        print(p)
        sys.stdout = sys.__stdout__  # reset printout to console

    def test_add_parameter(self):
        p = InferenceProblem("TestProblem")
        # check valid use cases for constant parameters
        p.add_parameter('c', 'model', const=1.0,
                        info='info', tex=r'$c$')
        p.add_parameter('loc_a', 'prior', const=1.0,
                        info='info', tex=r'$loc_a$')
        p.add_parameter('sigma_1', 'noise', const=1.0,
                        info='info', tex=r'$\sigma_1$')
        # check valid use cases for latent parameters
        p.add_parameter('b', 'model', prior=('normal', {'loc': 0, 'scale': 1}),
                        info='info', tex='$b$')
        p.add_parameter('scale_a', 'prior', info='info', tex='$b$',
                        prior=('uniform', {'low': 0.1, 'high': 2.0}))
        p.add_parameter('sigma_2', 'noise', info='info', tex=r'$\sigma_2$',
                        prior=('uniform', {'low': 0.1, 'high': 2.0}))
        p.add_parameter('a', 'model', info='info', tex='$a$',
                        prior=('normal', {'loc': 'loc_a', 'scale': 'scale_a'}))
        p.add_parameter('d', 'model')  # latent param. with uninformative prior
        # check invalid input arguments
        with self.assertRaises(RuntimeError):
            # adding a parameter with wrong type
            p.add_parameter('d', 'wrong_type', const=1.0)
        with self.assertRaises(RuntimeError):
            # adding a parameter with both const and prior given
            p.add_parameter('a', 'model', const=1.0,
                            prior=('uniform', {'low': 0.1, 'high': 2.0}))
        with self.assertRaises(RuntimeError):
            # adding a parameter with a name that was used before
            p.add_parameter('c', 'noise', const=2.0,
                            info='info', tex=r'$c$')
        with self.assertRaises(TypeError):
            # adding a parameter with an invalid prior
            p.add_parameter('eps', 'model', info='info', tex=r'$\epsilon$',
                            prior='Wrong prior format')
        with self.assertRaises(RuntimeError):
            # adding a parameter with an invalid prior
            p.add_parameter('eps', 'model', info='info', tex=r'$\epsilon$',
                            prior=[{'low': 0.1, 'high': 2.0}])
        with self.assertRaises(RuntimeError):
            # adding a parameter with an invalid prior
            p.add_parameter('eps', 'model', info='info', tex=r'$\epsilon$',
                            prior=('uniform', {'low': 0.1, 'high': 2.0}, '3'))
        with self.assertRaises(TypeError):
            # adding a parameter with an invalid prior
            p.add_parameter('eps', 'model', info='info', tex=r'$\epsilon$',
                            prior=(None, {'low': 0.1, 'high': 2.0}))
        with self.assertRaises(TypeError):
            # adding a parameter with an invalid prior
            p.add_parameter('eps', 'model', info='info', tex=r'$\epsilon$',
                            prior=('uniform', [{'low': 0.1}, {'high': 2.0}]))
        with self.assertRaises(TypeError):
            # adding a parameter with an invalid prior
            p.add_parameter('eps', 'model', info='info', tex=r'$\epsilon$',
                            prior=('uniform', {'low': None, 'high': 2.0}))

    def test_remove_parameter(self):
        # set up a problem with some parameters so we can remove them
        p = InferenceProblem("TestProblem")
        p.add_parameter('c', 'model', const=1.0,
                        info='info', tex=r'$c$')
        p.add_parameter('loc_a', 'prior', const=1.0,
                        info='info', tex=r'$loc_a$')
        p.add_parameter('s', 'noise', const=1.0,
                        info='info', tex=r'$\sigma$')
        p.add_parameter('scale_a', 'prior', info='info', tex='$b$',
                        prior=('uniform', {'low': 0.1, 'high': 2.0}))
        p.add_parameter('sigma', 'noise', info='info', tex=r'$\sigma$',
                        prior=('uniform', {'low': 0.1, 'high': 2.0}))
        p.add_parameter('a', 'model', info='info', tex='$a$',
                        prior=('normal', {'loc': 'loc_a', 'scale': 'scale_a'}))
        # check removing a constant parameter
        self.assertEqual(set(p.constant_prms),
                         {'c', 'loc_a', 's', 'low_scale_a', 'high_scale_a',
                          'low_sigma', 'high_sigma'})
        p.remove_parameter('c')  # <-- this is where the removal happens
        self.assertEqual(
            set(p.constant_prms), {'loc_a', 's', 'low_scale_a', 'high_scale_a',
                                   'low_sigma', 'high_sigma'})
        # check removing a latent parameter; note that removing a latent
        # parameter leads to the removal of all its prior parameters
        self.assertEqual(set(p.latent_prms), {'scale_a', 'sigma', 'a'})
        # check that indexes are as expected
        self.assertEqual(p.parameters['scale_a'].index, 0)
        self.assertEqual(p.parameters['sigma'].index, 1)
        self.assertEqual(p.parameters['a'].index, 2)
        p.remove_parameter('a')  # <-- this is where the removal happens
        # check that all latent parameters associated with 'a' are gone
        self.assertEqual(set(p.latent_prms), {'sigma'})
        # check if the re-indexing worked
        self.assertEqual(p.parameters['sigma'].index, 0)
        # check that all constant prior-parameters associated with 'a' are gone
        self.assertEqual(set(p.constant_prms), {'s', 'low_sigma', 'high_sigma'})
        # check that the removal of non-existing parameters results in an error
        with self.assertRaises(RuntimeError):
            p.remove_parameter('undefined')

    def test_check_if_parameter_exists(self):
        # check confirming existing parameter
        p = InferenceProblem("TestProblem")
        p.add_parameter('a', 'model', prior=('normal', {'loc': 0, 'scale': 1}))
        p.parameters.confirm_that_parameter_exists('a')
        # check RuntimeError for non-existing parameter
        with self.assertRaises(RuntimeError):
            p.parameters.confirm_that_parameter_exists('b')

    def test_change_parameter_role(self):
        p = InferenceProblem("TestProblem")
        # check change of role from latent to constant parameter
        p.add_parameter('a', 'model', prior=('normal', {'loc': 0, 'scale': 1}))
        self.assertEqual(set(p.latent_prms), {'a'})
        self.assertEqual(set(p.constant_prms), {'loc_a', 'scale_a'})
        self.assertEqual(set(p.prms), {'a', 'loc_a', 'scale_a'})
        p.change_parameter_role('a', const=1.0)  # <-- here the role changes
        self.assertEqual(set(p.latent_prms), set())
        self.assertEqual(set(p.constant_prms), {'a'})
        self.assertEqual(set(p.prms), {'a'})
        # check change of role from latent to constant parameter
        p.change_parameter_role('a', prior=('normal', {'loc': 0, 'scale': 1}))
        self.assertEqual(set(p.latent_prms), {'a'})
        self.assertEqual(set(p.constant_prms), {'loc_a', 'scale_a'})
        self.assertEqual(set(p.prms), {'a', 'loc_a', 'scale_a'})
        # check invalid input arguments
        with self.assertRaises(RuntimeError):
            # stated parameter does not exist
            p.change_parameter_role('undefined', const=1.0)
        with self.assertRaises(RuntimeError):
            # neither const nor prior are given
            p.change_parameter_role('a')
        with self.assertRaises(RuntimeError):
            # both const and prior are given
            p.change_parameter_role(
                'a', const=1.0, prior=('normal', {'loc': 0, 'scale': 1}))
        with self.assertRaises(RuntimeError):
            # change to role the parameter already has
            p.change_parameter_role(
                'a', prior=('normal', {'loc': 0, 'scale': 1}))

    def test_change_parameter_info(self):
        p = InferenceProblem("TestProblem")
        # simple check that the change works and has the expected effect
        p.add_parameter('a', 'model', prior=('normal', {'loc': 0, 'scale': 1}),
                        info='Info', tex='$a$')
        new_info, new_tex = 'New info', r'$\hat{a}$'
        p.change_parameter_info('a', new_info, new_tex=new_tex)
        self.assertEqual(p.parameters['a'].info, new_info)
        self.assertEqual(p.parameters['a'].tex, new_tex)
        # check call for non-existing parameter
        with self.assertRaises(RuntimeError):
            p.change_parameter_info('b', new_info, new_tex=new_tex)

    def test_change_constant(self):
        p = InferenceProblem("TestProblem")
        # simple check that the change works and has the expected effect
        p.add_parameter('a', 'model', const=0.0)
        p.add_parameter('b', 'model', prior=('normal', {'loc': 0, 'scale': 1}))
        new_const = 1.0
        p.change_constant('a', new_const)
        self.assertEqual(p.parameters['a'].value, new_const)
        # check for invalid input arguments
        with self.assertRaises(RuntimeError):
            # given parameter does not exist
            p.change_constant('undefined', 2.0)
        with self.assertRaises(RuntimeError):
            # given parameter is not a constant parameter
            p.change_constant('b', 2.0)

    def test_check_problem_consistency(self):
        p = InferenceProblem("TestProblem")
        # check if the method identifies inconsistent problems
        with self.assertRaises(AssertionError):
            # nothing is defined at this point
            p.check_problem_consistency()
        # define some parameters that do not require priors
        p.add_parameter('loc_a', 'prior', const=0.0)
        p.add_parameter('scale_a', 'prior', const=1.0)
        with self.assertRaises(AssertionError):
            # no priors defined yet
            p.check_problem_consistency()
        # add a prior by defining a latent parameter
        p.add_parameter('a', 'model', info='Some model parameter', tex='$a$',
                        prior=('normal', {'loc': 'loc_a', 'scale': 'scale_a'}))
        with self.assertRaises(AssertionError):
            # no forward models defined yet
            p.check_problem_consistency()
        # add a forward model
        test_model = ForwardModelBase('a', Sensor('x'), Sensor('y'))
        p.add_forward_model('TestModel', test_model)
        with self.assertRaises(AssertionError):
            # no noise models defined yet
            p.check_problem_consistency()
        # add a noise model
        p.add_parameter('s', 'noise', const=1.0)
        noise_model = NoiseModelBase('normal', 's', sensors=[Sensor('y')])
        p.add_noise_model(noise_model)
        with self.assertRaises(AssertionError):
            # no experiment_names defined yet
            p.check_problem_consistency()
        # add at an experiment
        p.add_experiment('Experiment_1', sensor_values={'x': 1, 'y': 1},
                         fwd_model_name='TestModel')
        # now the problem should be consistent
        p.check_problem_consistency()

    def test_add_experiment(self):
        # check correct use
        p = InferenceProblem("TestProblem")
        p.add_parameter('a', 'model', prior=('normal', {'loc': 0, 'scale': 1}))
        test_model = ForwardModelBase('a', Sensor('x'), Sensor('y'))
        p.add_forward_model('TestModel', test_model)
        p.add_experiment('Experiment_1', sensor_values={'x': 1, 'y': 1},
                         fwd_model_name='TestModel')
        # the experiment may contain more sensors than needed by forward model
        p.add_experiment('Experiment_2', sensor_values={'x': 1, 'y': 1, 'z': 1},
                         fwd_model_name='TestModel')
        # check that scalar sensor_values are not transformed to numpy arrays
        x_in_p = p.experiments['Experiment_2']['sensor_values']['x']
        self.assertEqual(type(x_in_p), type(1))
        # check invalid input arguments
        with self.assertRaises(TypeError):
            # wrong sensor_values type
            # noinspection PyTypeChecker
            p.add_experiment('Experiment_3', sensor_values=[('x', 1), ('y', 2)],
                             fwd_model_name='TestModel')
        with self.assertRaises(TypeError):
            # wrong fwd_model_name type
            # noinspection PyTypeChecker
            p.add_experiment('Experiment_3', sensor_values={'x': 1, 'y': 1},
                             fwd_model_name=1.2)
        with self.assertRaises(RuntimeError):
            # referencing non-existing forward model
            p.add_experiment('Experiment_3', sensor_values={'x': 1, 'y': 1},
                             fwd_model_name='ThisModelDoesNotExist')
        with self.assertRaises(RuntimeError):
            # forward model's input sensor not provided by experiment
            p.add_experiment('Experiment_3', sensor_values={'y': 1},
                             fwd_model_name='TestModel')
        with self.assertRaises(RuntimeError):
            # forward model's output sensor not provided by experiment
            p.add_experiment('Experiment_3', sensor_values={'x': 1},
                             fwd_model_name='TestModel')
        with self.assertRaises(RuntimeError):
            # sensor values with different lengths
            p.add_experiment('Experiment_3', fwd_model_name='TestModel',
                             sensor_values={'x': [1, 2], 'y': 1})
        # check that sensor_value lists are transformed to numpy arrays
        p.add_experiment('Experiment_3', fwd_model_name='TestModel',
                         sensor_values={'x': [1, 2], 'y': [2, 3]})
        x_in_p = p.experiments['Experiment_3']['sensor_values']['x']
        y_in_p = p.experiments['Experiment_3']['sensor_values']['y']
        self.assertEqual(type(x_in_p), np.ndarray)
        self.assertEqual(type(y_in_p), np.ndarray)
        # check adding the same experiment again
        p.add_experiment('Experiment_1', sensor_values={'x': 1, 'y': 1},
                         fwd_model_name='TestModel')

    def test_get_parameters(self):
        # check a simple use case
        p = InferenceProblem("TestProblem")
        p.add_parameter('c', 'model', const=1.0)
        p.add_parameter('a', 'model', prior=('normal', {'loc': 0, 'scale': 1}))
        p.add_parameter('b', 'model', prior=('normal', {'loc': 0, 'scale': 1}))
        a_value, b_value = 3.1, 14.7
        prms_def = {'b': 'b', 'a': 'a', 'loc_b': 'loc_b', 'c': 'c'}
        computed_result = p.get_parameters([a_value, b_value], prms_def)
        expected_result = {'b': b_value, 'a': a_value, 'loc_b': 0, 'c': 1.0}
        self.assertEqual(computed_result, expected_result)

    def test_get_experiment_names(self):
        # prepare for checks
        p = InferenceProblem("TestProblem")
        p.add_parameter('a', 'model', prior=('normal', {'loc': 0, 'scale': 1}))
        # define two forward models
        test_model_1 = ForwardModelBase('a', Sensor('x'),
                                        [Sensor('y1'), Sensor('y2')])
        test_model_2 = ForwardModelBase('a', Sensor('x'),
                                        [Sensor('z1'), Sensor('z2')])
        p.add_forward_model('TestModel_1', test_model_1)
        p.add_forward_model('TestModel_2', test_model_2)
        # define experiment_names to each forward model
        p.add_experiment('Experiment_Y1',
                         sensor_values={'x': 1, 'y1': 1, 'y2': 1},
                         fwd_model_name='TestModel_1')
        p.add_experiment('Experiment_Y2',
                         sensor_values={'x': 2, 'y1': 2, 'y2': 2},
                         fwd_model_name='TestModel_1')
        p.add_experiment('Experiment_Z1',
                         sensor_values={'x': -1, 'z1': -1, 'z2': -1},
                         fwd_model_name='TestModel_2')
        p.add_experiment('Experiment_Z2',
                         sensor_values={'x': -2, 'z1': -2, 'z2': -2},
                         fwd_model_name='TestModel_2')
        # get experiment_names for specific forward model
        computed_result = p.get_experiment_names(
            forward_model_names='TestModel_1')
        expected_result = ['Experiment_Y1', 'Experiment_Y2']
        for i in range(len(computed_result)):
            self.assertEqual(computed_result[i], expected_result[i])
        # get experiment_names for specific forward model
        computed_result = p.get_experiment_names(
            forward_model_names='TestModel_2')
        expected_result = ['Experiment_Z1', 'Experiment_Z2']
        for i in range(len(computed_result)):
            self.assertEqual(computed_result[i], expected_result[i])
        # get experiment_names for both forward models
        computed_result = p.get_experiment_names(
            forward_model_names=['TestModel_1', 'TestModel_2'])
        expected_result = ['Experiment_Y1', 'Experiment_Y2',
                           'Experiment_Z1', 'Experiment_Z2']
        for i in range(len(computed_result)):
            self.assertEqual(computed_result[i], expected_result[i])
        # get experiment_names via output sensors
        computed_result = p.get_experiment_names(sensor_names=['y1', 'y2'])
        expected_result = ['Experiment_Y1', 'Experiment_Y2']
        for i in range(len(computed_result)):
            self.assertEqual(computed_result[i], expected_result[i])
        # get experiment_names via output sensors
        computed_result = p.get_experiment_names(sensor_names=['z1', 'z2'])
        expected_result = ['Experiment_Z1', 'Experiment_Z2']
        for i in range(len(computed_result)):
            self.assertEqual(computed_result[i], expected_result[i])
        # get experiment_names via output sensors
        computed_result = p.get_experiment_names(
            sensor_names=['z1', 'y1', 'z2'])
        expected_result = []
        self.assertEqual(computed_result, expected_result)
        # get experiment_names via both forward model name and sensors
        computed_result = p.get_experiment_names(
            sensor_names=['y1', 'y2'], forward_model_names='TestModel_1')
        expected_result = ['Experiment_Y1', 'Experiment_Y2']
        for i in range(len(computed_result)):
            self.assertEqual(computed_result[i], expected_result[i])
        # get experiment_names via both forward model name and sensors
        computed_result = p.get_experiment_names(
            sensor_names=['y1', 'y2'], forward_model_names='TestModel_2')
        expected_result = []
        self.assertEqual(computed_result, expected_result)
        # check invalid input arguments
        with self.assertRaises(RuntimeError):
            # neither forward_model_names nor sensor_names given
            p.get_experiment_names()

    def test_get_theta_names(self):
        # check some simple use cases
        p = InferenceProblem("TestProblem")
        p.add_parameter(
            'a', 'model', prior=('normal', {'loc': 0, 'scale': 1}), tex="$a$")
        p.add_parameter(
            'b', 'model', prior=('normal', {'loc': 0, 'scale': 1}), tex="$b$")
        computed_result = p.get_theta_names(tex=False)
        expected_result = ['a', 'b']
        self.assertEqual(computed_result, expected_result)
        computed_result = p.get_theta_names(tex=True)
        expected_result = ['$a$', '$b$']
        self.assertEqual(computed_result, expected_result)

    def test_theta_explanation(self):
        # simply check that no errors occur when theta_explanation is called;
        # the output that is usually printed is redirected
        p = InferenceProblem("TestProblem")
        p.add_parameter('a', 'model', const=1.0)
        p.add_parameter('b', 'model', prior=('normal', {'loc': 0, 'scale': 1}))
        p.add_parameter('s', 'noise', prior=('normal', {'loc': 0, 'scale': 1}))
        # check the model_consistency flag
        with self.assertRaises(AssertionError):
            # the model is not consistent
            p.theta_explanation(check_consistency=True)

    def test_add_forward_model(self):
        # check correct use
        p = InferenceProblem("TestProblem")
        p.add_parameter('a', 'model', prior=('normal', {'loc': 0, 'scale': 1}))
        test_model = ForwardModelBase('a', Sensor('x'), Sensor('y'))
        p.add_forward_model('TestModel', test_model)
        # check for invalid input arguments
        with self.assertRaises(RuntimeError):
            # a given parameter of the forward model has not been added yet
            test_model_2 = ForwardModelBase('b', Sensor('x'), Sensor('y'))
            p.add_forward_model('TestModel_2', test_model_2)
        with self.assertRaises(RuntimeError):
            # add a forward model with the same name
            p.add_forward_model('TestModel', test_model)
        with self.assertRaises(RuntimeError):
            # add a forward model with already used output sensor name
            test_model_2 = ForwardModelBase('a', Sensor('x'), Sensor('y'))
            p.add_forward_model('TestModel_2', test_model_2)

    def test_evaluate_model_response(self):
        # prepare for checks
        p = InferenceProblem("TestProblem")
        p.add_parameter('a0', 'model', prior=('normal', {'loc': 0, 'scale': 1}))
        p.add_parameter('a1', 'model', prior=('normal', {'loc': 0, 'scale': 1}))
        p.add_parameter('a2', 'model', prior=('normal', {'loc': 0, 'scale': 1}))

        class FwdModel(ForwardModelBase):
            def __call__(self, inp):
                x = inp['x']
                a0 = inp['a0']
                a1 = inp['a1']
                a2 = inp['a2']
                return {'y': a0 + a1 * x + a2 * x ** 2}

        # add forward model
        fwd_model = FwdModel(['a0', 'a1', 'a2'], Sensor('x'), Sensor('y'))
        p.add_forward_model('FwdModel', fwd_model)
        # add experiment_names
        p.add_experiment('Exp1', sensor_values={'x': 1, 'y': 2},
                         fwd_model_name='FwdModel')
        p.add_experiment('Exp2', sensor_values={'x': 2, 'y': 3},
                         fwd_model_name='FwdModel')
        p.add_experiment('Exp3', sensor_values={'x': [1, 2], 'y': [1, 2]},
                         fwd_model_name='FwdModel')
        # perform a check for all experiments
        a0_value, a1_value, a2_value = 1, 2, 3
        theta = np.array([a0_value, a1_value, a2_value])
        computed_result = p.evaluate_model_response(theta)
        expected_result = {'Exp1': {'y': 6},
                           'Exp2': {'y': 17},
                           'Exp3': {'y': np.array([6, 17])}}
        # check for each item, because the last contains an array, which results
        # in an error is you do assertEqual on the whole thing
        self.assertEqual(computed_result['Exp1'], expected_result['Exp1'])
        self.assertEqual(computed_result['Exp2'], expected_result['Exp2'])
        self.assertEqual(computed_result['Exp3']['y'][0],
                         expected_result['Exp3']['y'][0])
        self.assertEqual(computed_result['Exp3']['y'][1],
                         expected_result['Exp3']['y'][1])
        # perform a check for a subset of experiments
        computed_result = p.evaluate_model_response(
            theta, experiment_names='Exp3')
        expected_result = {'Exp3': {'y': np.array([6, 17])}}
        self.assertEqual(computed_result['Exp3']['y'][0],
                         expected_result['Exp3']['y'][0])
        self.assertEqual(computed_result['Exp3']['y'][1],
                         expected_result['Exp3']['y'][1])

    def test_add_noise_model(self):
        # check correct use
        p = InferenceProblem("TestProblem")
        p.add_parameter('a', 'model', prior=('normal', {'loc': 0, 'scale': 1}))
        p.add_parameter('s1', 'noise', prior=('normal', {'loc': 0, 'scale': 1}))
        p.add_parameter('s2', 'noise', prior=('normal', {'loc': 0, 'scale': 1}))
        p.add_parameter('s3', 'noise', prior=('normal', {'loc': 0, 'scale': 1}))
        test_model = ForwardModelBase('a', Sensor('x'),
                                      [Sensor('y1'), Sensor('y2')])
        p.add_forward_model('TestModel', test_model)
        noise_model1 = NoiseModelBase('normal', 's1',
                                      sensors=[Sensor('y1'), Sensor('y2')])
        noise_model2 = NoiseModelBase('normal', ['s2'], sensors=[Sensor('y2')])
        noise_model3 = NoiseModelBase('normal', ['s1', 's2', 's3'],
                                      sensors=[Sensor('y1')])
        p.add_noise_model(noise_model1)
        p.add_noise_model(noise_model2)
        p.add_noise_model(noise_model3)
        # check invalid input arguments
        with self.assertRaises(RuntimeError):
            # the given noise model parameter has not been defined
            p.add_noise_model(NoiseModelBase('normal', 'not_existing_parameter',
                                             sensors=[Sensor('y1'),
                                                      Sensor('y2')]))

    def test_assign_experiments_to_noise_models(self):
        # some preparations before performing the actual tests
        p = InferenceProblem("TestProblem")
        # add some parameters
        p.add_parameter('a', 'model', prior=('normal', {'loc': 0, 'scale': 1}))
        p.add_parameter('s1', 'noise', prior=('normal', {'loc': 0, 'scale': 1}))
        p.add_parameter('s2', 'noise', prior=('normal', {'loc': 0, 'scale': 1}))
        p.add_parameter('s3', 'noise', prior=('normal', {'loc': 0, 'scale': 1}))
        # add some dummy forward models (only the output sensors are important)
        test_model_y1 = ForwardModelBase('a', Sensor('x'), Sensor('y1'))
        test_model_y2 = ForwardModelBase('a', Sensor('x'), Sensor('y2'))
        test_model_z1z2 = ForwardModelBase('a', Sensor('x'),
                                           [Sensor('z1'), Sensor('z2')])
        p.add_forward_model('TestModel_y1', test_model_y1)
        p.add_forward_model('TestModel_y2', test_model_y2)
        p.add_forward_model('TestModel_z1z2', test_model_z1z2)
        # add some noise models
        noise_model_y1 = NoiseModelBase('normal', ['s1', 's2', 's3'],
                                        sensors=[Sensor('y1')])
        noise_model_y2 = NoiseModelBase(
            'normal', ['s2'], sensors=[Sensor('y2')])
        noise_model_y1y2 = NoiseModelBase(
            'normal', 's1', sensors=[Sensor('z1'), Sensor('z2')])
        p.add_noise_model(noise_model_y1)
        p.add_noise_model(noise_model_y2)
        p.add_noise_model(noise_model_y1y2)
        # add some experiments
        p.add_experiment('Exp_y1', sensor_values={'x': 1, 'y1': 2},
                         fwd_model_name='TestModel_y1')
        p.add_experiment('Exp_y2_1', sensor_values={'x': 3, 'y2': 4},
                         fwd_model_name='TestModel_y2')
        p.add_experiment('Exp_y2_2', sensor_values={'x': 5, 'y2': 6},
                         fwd_model_name='TestModel_y2')
        p.add_experiment('Exp_z1z2', sensor_values={'x': 7, 'z1': 8, 'z2': 9},
                         fwd_model_name='TestModel_z1z2')
        # this is the call that should be tested here
        p.assign_experiments_to_noise_models()
        # now check if all experiments have been assigned correctly
        self.assertEqual(
            p.noise_models[0].experiment_names, ['Exp_y1'])
        self.assertEqual(
            p.noise_models[1].experiment_names,  ['Exp_y2_1', 'Exp_y2_2'])
        self.assertEqual(
            p.noise_models[2].experiment_names, ['Exp_z1z2'])

    def test_transform_experimental_data(self):
        # check correct use
        p = InferenceProblem("TestProblem")
        p.add_parameter('a', 'model', prior=('normal', {'loc': 0, 'scale': 1}))
        forward_model = ForwardModelBase('a', Sensor('x'), Sensor('y'))
        p.add_forward_model('TestModel', forward_model)
        p.add_experiment('Experiment_1', fwd_model_name='TestModel',
                         sensor_values={'x': [1, 2], 'y': [1, 3]})
        # apply a simple power function to the experimental data
        p_copy = p.transform_experimental_data(
            f=np.power, args=([2, 3],), where=True)
        x_computed = p_copy.experiments['Experiment_1']['sensor_values']['x']
        x_expected = np.array([1, 8])
        self.assertTrue(np.allclose(x_computed, x_expected) and
                                    x_computed.shape == x_expected.shape)
        y_computed = p_copy.experiments['Experiment_1']['sensor_values']['y']
        y_expected = np.array([1, 27])
        self.assertTrue(np.allclose(y_computed, y_expected) and
                        y_computed.shape == y_expected.shape)

if __name__ == "__main__":
    unittest.main()



# standard library imports
import time
import random
import contextlib

# third party imports
import numpy as np
import emcee
import arviz as az
from loguru import logger
from tabulate import tabulate

# local imports
from probeye.subroutines import pretty_time_delta
from probeye.inference.scipy_.solver import ScipySolver
from probeye.subroutines import stream_to_logger, print_dict_in_rows


class EmceeSolver(ScipySolver):
    """
    Provides emcee-sampler which is a pure-Python implementation of Goodman &
    Weare’s Affine Invariant Markov chain Monte Carlo (MCMC) Ensemble sampler,
    see https://emcee.readthedocs.io/en/stable/.
    """

    def __init__(self, problem, seed=1, show_progress=True):
        """See docstring of ScipySolver for information on the arguments."""
        logger.debug("Initializing EmceeSolver")
        # initialize the scipy-based solver (ScipySolver)
        super().__init__(problem, seed=seed, show_progress=show_progress)

    def emcee_summary(self, posterior_samples):
        """
        Computes and prints a summary of the posterior samples containing
        mean, median, standard deviation, 5th percentile and 95th percentile.
        Note, that this method was based on code from the taralli package, see
        https://gitlab.com/tno-bim/taralli.

        Parameters
        ----------
        posterior_samples : numpy.ndarray
            The generated samples in an array with as many columns as there are
            latent parameters, and n rows, where n = n_chains * n_steps.

        Returns
        -------
        dict
            Keys are the different statistics 'mean', 'median', 'sd' (standard
            deviation), 'q05' and 'q95' (0.05- and 0.95-quantile). The values
            are dictionaries with the parameter names as keys and the respective
            statistic-values as values.
        """

        # used for the names in the first column
        var_names = self.problem.get_theta_names(tex=False)

        # compute some stats for each column (i.e., each parameter)
        mean = np.mean(posterior_samples, axis=0)
        quantiles = np.quantile(posterior_samples, [0.50, 0.05, 0.95], axis=0)
        median = quantiles[0, :]
        quantile_05 = quantiles[1, :]
        quantile_95 = quantiles[2, :]

        # compute the sample standard deviations for each parameter
        cov_matrix = np.atleast_2d(np.cov(posterior_samples.T))
        sd = np.sqrt(np.diag(cov_matrix))

        # assemble the summary array
        col_names = ["", "mean", "median", "sd", "5%", "95%"]
        row_names = np.array(var_names).reshape(-1, 1)
        tab = np.hstack((row_names,
                         mean.reshape(-1, 1),
                         median.reshape(-1, 1),
                         sd.reshape(-1, 1),
                         quantile_05.reshape(-1, 1),
                         quantile_95.reshape(-1, 1)))

        # print the generated table, and return a summary dict for later use
        print(tabulate(tab, headers=col_names, floatfmt=".2f"))
        return {'mean': {name: val for name, val in zip(mean, row_names)},
                'median': {name: val for name, val in zip(median, row_names)},
                'sd':  {name: val for name, val in zip(sd, row_names)},
                'q05': {name: val for name, val in zip(quantile_05, row_names)},
                'q95': {name: val for name, val in zip(quantile_95, row_names)}}

    def run_mcmc(self, n_walkers=20, n_steps=1000, n_initial_steps=100,
                 **kwargs):
        """
        Runs the emcee-sampler for the InferenceProblem the EmceeSolver was
        initialized with and returns the results as an arviz InferenceData obj.

        Parameters
        ----------
        n_walkers : int, optional
            Number of walkers used by the estimator.
        n_steps : int, optional
            Number of steps to run.
        n_initial_steps : int, optional
            Number of steps for initial (burn-in) sampling.
        **kwargs : optional
            Additional key-word arguments channeled to emcee.EnsembleSampler.

        Returns
        -------
        inference_data : obj[arviz.data.inference_data.InferenceData]
            Contains the results of the sampling procedure.
        """

        # log which solver is used
        logger.info(
            f"Solving problem using emcee sampler with {n_initial_steps} + "
            f"{n_steps} samples and {n_walkers} walkers")
        if kwargs:
            logger.info("Additional options:")
            print_dict_in_rows(kwargs, printer=logger.info)
        else:
            logger.info("No additional options specified")

        # draw initial samples from the parameter's priors
        logger.debug("Drawing initial samples")
        sampling_initial_positions = np.zeros(
            (n_walkers, self.problem.n_latent_prms))
        theta_names = self.problem.get_theta_names()
        for i, parameter_name in enumerate(theta_names):
            sampling_initial_positions[:, i] = self.sample_from_prior(
                parameter_name, n_walkers)

        # The following code is based on taralli and merely adjusted to the
        # variables in the probeye setup; see https://gitlab.com/tno-bim/taralli

        # .................................................................... #
        #                             Pre-process                              #
        # .................................................................... #

        n_rows, n_cols = sampling_initial_positions.shape
        n_latent_prms = self.problem.n_latent_prms

        if n_cols != self.problem.n_latent_prms:
            raise ValueError(
                f"'sampling_initial_positions' should have {n_latent_prms} "
                f"columns (one for each latent parameter), but {n_cols} are "
                f"provided.")

        if n_rows != n_walkers:
            raise ValueError(
                f"'sampling_initial_positions' should have {n_walkers} rows "
                f"(one for each walker), but {n_rows} are provided.")

        random.seed(self.seed)
        np.random.seed(self.seed)
        rstate = np.random.mtrand.RandomState(self.seed)

        logger.debug("Setting up EnsembleSampler")
        sampler = emcee.EnsembleSampler(
            nwalkers=n_walkers,
            ndim=self.problem.n_latent_prms,
            log_prob_fn=lambda x: self.logprior(x) + self.loglike(x),
            **kwargs)

        sampler.random_state = rstate

        # .................................................................... #
        #    Initial sampling, burn-in: used to avoid a poor starting point    #
        # .................................................................... #

        logger.debug("Starting sampling (initial + main)")
        start = time.time()
        state = sampler.run_mcmc(
            initial_state=sampling_initial_positions,
            nsteps=n_initial_steps,
            progress=self.show_progress)
        sampler.reset()

        # .................................................................... #
        #                      Sampling of the posterior                       #
        # .................................................................... #
        sampler.run_mcmc(
            initial_state=state, nsteps=n_steps, progress=self.show_progress)
        end = time.time()
        runtime_str = pretty_time_delta(end - start)
        logger.info(
            f"Sampling of the posterior distribution completed: {n_steps} steps"
            f" and {n_walkers} walkers.")
        logger.info(
            f"Total run-time (including initial sampling): {runtime_str}.")
        logger.info("")
        logger.info("Summary of sampling results")
        posterior_samples = sampler.get_chain(flat=True)
        with contextlib.redirect_stdout(stream_to_logger('INFO')):
            self.emcee_summary(posterior_samples)
        self.raw_results = sampler

        # translate the results to a common data structure and return it
        var_names = self.problem.get_theta_names(tex=True)
        inference_data = az.from_emcee(sampler, var_names=var_names)
        return inference_data



# module imports
from probeye.inference.scipy_ import correlation_models
from probeye.inference.scipy_ import noise_models
from probeye.inference.scipy_ import priors
from probeye.inference.scipy_ import solver



# standard library imports
import unittest

# third party imports
import numpy as np

# local imports
from probeye.definition.forward_model import ForwardModelBase
from probeye.definition.sensor import Sensor

class TestProblem(unittest.TestCase):

    def test_undefined_forward_model(self):
        # check for the not-implemented-error when no response method is defined
        forward_model = ForwardModelBase(['a'], Sensor('x'), Sensor('y'))
        with self.assertRaises(NotImplementedError):
            forward_model.response({})
        with self.assertRaises(NotImplementedError):
            forward_model({})

    def test_model_template_one_sensor(self):

        # define a simple model using ForwardModelBase
        class ForwardModel(ForwardModelBase):
            def response(self, inp):
                x = inp['x']
                a = inp['a']
                b = inp['b']
                return {'y': a * x ** 2 + b}

        # check the __call__-method
        forward_model = ForwardModel(
            ['a', 'b'], [Sensor('x')], [Sensor('y')])
        prms = {'a': 1, 'b': 2}
        computed_result = forward_model({**{'x': 1.0}, **prms})
        expected_result = {'y': 3.0}
        self.assertEqual(computed_result, expected_result)

        # check input/output sensor names
        self.assertEqual(forward_model.input_sensor_names, ['x'])
        self.assertEqual(forward_model.output_sensor_names, ['y'])

        # check the jacobian-method (dict-version)
        computed_result = forward_model.jacobian({**{'x': 1.0}, **prms})
        expected_result = {'y': {'x': 2.0, 'a': 1.0, 'b': 1.0}}
        for k1, v1 in computed_result.items():
            for k2, v2 in v1.items():
                self.assertAlmostEqual(v2, expected_result[k1][k2], places=2)
        # check the jacobian-method (array-version)
        inp_ = {**{'x': 1.0}, **prms}
        jac_dict = forward_model.jacobian(inp_)
        computed_result = forward_model.jacobian_dict_to_array(inp_, jac_dict)
        expected_result = np.array([[2.0, 1.0, 1.0]])
        self.assertTrue(np.allclose(computed_result, expected_result, atol=1e-3)
                        and computed_result.shape == expected_result.shape)

    def test_model_template_multiple_sensors(self):

        # define an output sensor with an offset attribute
        class OutputSensorOffset(Sensor):
            def __init__(self, name, offset):
                super().__init__(name)
                self.offset = offset

        # define a simple model using ForwardModelBase
        class ForwardModel(ForwardModelBase):
            def response(self, inp):
                x1 = inp['x1']
                x2 = inp['x2']
                a = inp['a']
                b = inp['b']
                response_dict = dict()
                for os in self.output_sensors:
                    response_dict[os.name] = a * x1**2 + b * x2 + os.offset
                return response_dict

        # check the __call__-method
        is1, is2 = Sensor('x1'), Sensor('x2')
        os1, os2 = OutputSensorOffset('y1', 0.1), OutputSensorOffset('y2', -0.2)
        forward_model = ForwardModel(['a', 'b'], [is1, is2], [os1, os2])
        prms = {'a': 1, 'b': 2}
        computed_result = forward_model({**{'x1': 2.0, 'x2': 3.0}, **prms})
        expected_result = {'y1': 10.1, 'y2': 9.8}
        self.assertEqual(computed_result, expected_result)

        # check input/output sensor names
        self.assertEqual(forward_model.input_sensor_names, ['x1', 'x2'])
        self.assertEqual(forward_model.output_sensor_names, ['y1', 'y2'])

        # check the jacobian-method (dict-version)
        computed_result = forward_model.jacobian(
            {**{'x1': 2.0, 'x2': 3.0}, **prms})
        expected_result = {'y1': {'x1': 4.0, 'x2': 2.0, 'a': 4.0, 'b': 3.0},
                           'y2': {'x1': 4.0, 'x2': 2.0, 'a': 4.0, 'b': 3.0}}
        for k1, v1 in computed_result.items():
            for k2, v2 in v1.items():
                self.assertAlmostEqual(v2, expected_result[k1][k2], places=2)
        # check the jacobian-method (array-version)
        inp_ = {**{'x1': 2.0, 'x2': 3.0}, **prms}
        jac_dict = forward_model.jacobian(inp_)
        computed_result = forward_model.jacobian_dict_to_array(inp_, jac_dict)
        expected_result = np.array([[4.0, 2.0, 4.0, 3.0],
                                    [4.0, 2.0, 4.0, 3.0]])
        self.assertTrue(
            np.allclose(computed_result, expected_result, atol=1e-3) and
            computed_result.shape == expected_result.shape)

if __name__ == "__main__":
    unittest.main()



# standard library imports
import copy as cp

# third party imports
import copy
import numpy as np
from scipy.optimize import minimize
from loguru import logger

# local imports
from probeye.inference.scipy_.priors import translate_prior
from probeye.inference.scipy_.noise_models import translate_noise_model
from probeye.subroutines import print_dict_in_rows


class ScipySolver:
    """Solver routines based on scipy and numpy for an InferenceProblem."""

    def __init__(self, problem, seed=1, show_progress=True):
        """
        Parameters
        ----------
        problem : obj[InferenceProblem]
            Describes the inference problem including e.g. parameters and data.
        seed : int, optional
            Random state used for random number generation.
        show_progress : bool, optional
            When True, the progress of a solver routine will be shown (for
            example as a progress-bar) if such a feature is available.
            Otherwise, the progress will not shown.
        """

        # log at beginning so that errors can be associated
        logger.debug("Initializing ScipySolver")

        # attributes from arguments
        self.show_progress = show_progress
        self.seed = seed

        # the following attribute will be set after the solver was run
        self.raw_results = None

        # each noise model must be connected to the relevant experiment_names;
        # a copy is created before, so that this solver routine does not have
        # side effects on the original problem; such side effects would occur
        # due to calling the assign_experiments_to_noise_models-method below
        self.problem = cp.deepcopy(problem)
        self.problem.assign_experiments_to_noise_models()

        # translate the prior definitions to objects with computing capabilities
        logger.debug("Translate problem's priors")
        self.priors = copy.deepcopy(self.problem.priors)
        for prior_name, prior_template in self.problem.priors.items():
            self.priors[prior_name] = translate_prior(prior_template)

        # translate the general noise model objects into solver specific ones
        logger.debug("Translate problem's noise models")
        self.noise_models = []
        for noise_model_base in self.problem.noise_models:
            self.noise_models.append(translate_noise_model(noise_model_base))

    def logprior(self, theta):
        """
        Evaluates the log-prior function of the problem at theta.

        Parameters
        ----------
        theta : array_like
            A numeric vector for which the log-likelihood function should be
            evaluated. Which parameters these numbers refer to can be checked
            by calling self.theta_explanation() once the problem is set up.

        Returns
        -------
        lp : float
            The evaluated log-prior function for the given theta-vector.
        """
        lp = 0.0
        for prior in self.priors.values():
            prms = self.problem.get_parameters(theta, prior.prms_def)
            lp += prior(prms, 'logpdf')
        return lp

    def sample_from_prior(self, prm_name, size):
        """
        Generates random samples from a parameter's prior distribution and
        returns the generated samples.

        Parameters
        ----------
        prm_name : string
            The name of the parameter the prior is associated with.
        size : int
            The number of random samples to be drawn.

        Returns
        -------
        numpy.ndarray
            The generated samples.
        """
        prior = self.priors[self.problem.parameters[prm_name].prior.name]
        # check for prior-priors; if a prior parameter is a latent
        # parameter and not a constant, one first samples from the prior
        # parameter's prior distribution, and then takes the mean of those
        # samples to sample from the first prior distribution; this procedure
        # is recursive, so that (in principle) one could also define priors of
        # the prior's prior parameters and so forth
        theta_aux = [0] * self.problem.parameters.n_latent_prms
        for prior_prm_name in prior.hyperparameters.keys():
            if self.problem.parameters[prior_prm_name].role == 'latent':
                samples = self.sample_from_prior(prior_prm_name, size)
                theta_aux[self.problem.parameters[prior_prm_name].index] =\
                    np.mean(samples)
        prms = self.problem.get_parameters(theta_aux, prior.hyperparameters)
        return prior.generate_samples(prms, size)

    def loglike(self, theta):
        """
        Evaluates the log-likelihood function of the problem at theta.

        Parameters
        ----------
        theta : array_like
            A numeric vector for which the log-likelihood function should be
            evaluated. Which parameters these numbers refer to can be checked
            by calling self.theta_explanation() once the problem is set up.

        Returns
        -------
        ll : float
            The evaluated log-likelihood function for the given theta-vector.
        """
        # compute the contribution to the log-likelihood function for each noise
        # model and sum it all up
        ll = 0.0
        for noise_model in self.noise_models:
            # compute the model response for the noise model's experiment_names
            model_response = self.problem.evaluate_model_response(
                theta, noise_model.experiment_names)
            # get the parameter values for the noise model's parameters
            prms_noise = self.problem.get_parameters(
                theta, noise_model.prms_def)
            # evaluate the loglike-contribution for the noise model
            ll += noise_model.loglike_contribution(model_response, prms_noise)
        return ll

    def get_start_values(self, x0_dict, x0_prior, x0_default):
        """
        Derives the start values for the maximum likelihood optimization run.
        For an explanation of the arguments, see self.run_max_likelihood.

        Returns
        -------
        tuple
            x0 : numpy.ndarray
                A numeric vector with the derived start values in the order of
                InferenceProblem.get_theta_names().
            x0_dict : dict
                Keys are the latent parameters, while the keys are their start
                values.
        """

        # this is going to be the start vector
        x0 = np.zeros(self.problem.n_latent_prms)

        if x0_dict:
            # in this case, the user explicitly defined the start values
            for prm_name, prm_value in x0_dict.items():
                idx = self.problem.parameters[prm_name].index
                x0[idx] = prm_value
        else:
            # in this case, the start values are derived from the priors; if
            # a prior i not uninformative, its mean value will be used; if a
            # prior is uninformative, the x0_default value will be used
            x0_dict = {}
            prms = cp.copy(self.problem.constant_prms_dict)
            for prm_name in self.problem.get_theta_names():
                prior_name = self.problem.parameters[prm_name].prior.name
                prior_type = self.problem.parameters[prm_name].prior.prior_type
                idx = self.problem.parameters[prm_name].index
                if prior_type != 'uninformative':
                    prm_value = self.priors[prior_name](
                        prms, x0_prior, use_ref_prm=False)
                    prms[prm_name] = prm_value
                    x0[idx] = prm_value
                else:
                    # no mean value can be requested if the prior is
                    # uninformative, hence a default value is used
                    x0[idx] = x0_default
                x0_dict[prm_name] = x0[idx]

        return x0, x0_dict

    def summarize_ml_results(self, minimize_results, true_values, x0_dict):
        """
        Prints a summary of the results of the maximum likelihood estimation.
        For an explanation of the arguments, check out the docstring of the
        self.run_max_likelihood-method.
        """

        # the first part of the summary contains process information
        n_char_message = len(minimize_results.message)
        msg = (f"\nResults of maximum likelihood estimation\n"
               f"{'=' * n_char_message}\n"
               f"{minimize_results.message}\n"
               f"{'-' * n_char_message}\n"
               f"Number of iterations:           {minimize_results.nit}\n"
               f"Number of function evaluations: {minimize_results.nfev}\n"
               f"{'-' * n_char_message}")
        for line in msg.split('\n'):
            logger.info(line)

        # the second part shows the actual results and compares them with the
        # true values (if given) and the start values
        if minimize_results.success:
            theta_names = self.problem.get_theta_names(tex=False)
            n_char = max([len(name) for name in theta_names]) + 4
            for i, theta_name in enumerate(theta_names):
                opt_name = f"{theta_name}_opt"
                line = f"{opt_name:{n_char}s} = {minimize_results.x[i]:.6f}"
                if true_values:
                    line += (f" (true = {true_values[theta_name]}, "
                             f"start = {x0_dict[theta_name]})")
                else:
                    line += f" (start = {x0_dict[theta_name]})"
                logger.info(line)
        logger.info('')  # empty line for visual buffer

    def run_max_likelihood(self, x0_dict=None, x0_prior='mean', x0_default=1.0,
                           true_values=None, method='Nelder-Mead',
                           solver_options=None):
        """
        Finds values for an InferenceProblem's latent parameters that maximize
        the problem's likelihood function. The used method is scipy's minimize
        function from the optimize submodule.

        Parameters
        ----------
        x0_dict : dict, optional
            Contains the start values for each latent variable. Via this arg
            the user can explicitly specify a start value for the optimization.
        x0_prior : {'mean', 'median'}, optional
            If x0_dict is not given, the start values will be derived from the
            priors, either using the 'mean' or 'median' value. If x0_dict is
            given, this argument has no effect.
        x0_default : float, optional
            For uninformative priors, no mean or median value is defined. In
            those cases, the default_x0 value will be used as start value. If
            x0_dict is given, this argument has no effect.
        true_values : None, dict, optional
            Defines 'true' parameter values. Keys are the parameter names and
            values are the 'true' values. They are only used to print them next
            to the inferred parameter values from the optimization run.
        method : str, optional
            Defines the algorithm used by scipy.optimize.minimize. See the
            documentation of this scipy method to see all the options.
        solver_options : dict, optional
            Options passed to scipy.optimize.minimize under the 'options' key
            word argument. See the documentation of this scipy method to see
            available options.

        Returns
        -------
        minimize_results : obj[scipy.optimize.optimize.OptimizeResult]
            An object returns by scipy's minimize function containing the
            optimization results. The parameter vector that optimizes the
            likelihood function can be requested via 'minimize_results.x'.
        """

        # log at beginning so that errors can be associated
        logger.info("Solving problem via maximum likelihood estimation")

        # since scipy's minimize function is used, we need a function that
        # returns the negative log-likelihood function (minimizing the negative
        # log-likelihood is equivalent to maximizing the (log-)likelihood)
        def fun(x):
            return -self.loglike(x)

        # prepare the start value either from the given x0_dict or from the mean
        # values of the latent parameter's priors
        logger.debug("Deriving start values")
        x0, x0_dict = self.get_start_values(x0_dict, x0_prior, x0_default)
        logger.info("Using start values:")
        print_dict_in_rows(x0_dict, printer=logger.info, val_fmt='.2f')

        # this is the where the solver does its thing
        logger.info(f"Starting optimizer (using {method})")
        if solver_options:
            logger.info("Specified solver options:")
            print_dict_in_rows(solver_options, printer=logger.info)
        else:
            logger.info("No solver options specified")
        minimize_results = minimize(
            fun, x0, method=method, options=solver_options)

        # note that in this case, the raw solver result is identical with the
        # return-value of this method; however, for other solver they differ;
        # hence, this attribute is set here only for consistency reasons
        self.raw_results = minimize_results

        # some convenient printout with respect to the solver's results
        self.summarize_ml_results(minimize_results, true_values, x0_dict)

        return minimize_results



# standard library imports
import unittest

# local imports
from probeye.definition.prior import PriorBase


class TestProblem(unittest.TestCase):

    def test_prior_template(self):
        prior_template = PriorBase('a', ['loc_a', 'scale_a'],
                                       'a_normal', 'normal distribution')
        # check that the attributes are wired correctly
        self.assertEqual(prior_template.ref_prm, 'a')
        self.assertEqual(prior_template.name, 'a_normal')
        self.assertEqual(prior_template.prior_type, 'normal distribution')
        self.assertEqual(prior_template.prms_def,
                         {'a': 'a', 'loc_a': 'loc_a', 'scale_a': 'scale_a'})
        self.assertEqual(prior_template.hyperparameters,
                         {'loc_a': 'loc_a', 'scale_a': 'scale_a'})
        # check that the __str__-method works
        self.assertTrue(len(prior_template.__str__()) > 0)

if __name__ == "__main__":
    unittest.main()



"""
Simple linear regression example with two model and one noise parameter
--------------------------------------------------------------------------------
The model equation is y = a * x + b with a, b being the model parameters and the
noise model is a normal zero-mean distribution with the std. deviation to infer.
The problem is solved via sampling using emcee and pyro.
"""

# standard library imports
import unittest

# third party imports
import numpy as np
import matplotlib.pyplot as plt

# local imports (problem definition)
from probeye.definition.inference_problem import InferenceProblem
from probeye.definition.forward_model import ForwardModelBase
from probeye.definition.sensor import Sensor
from probeye.definition.noise_model import NormalNoiseModel

# local imports (testing related)
from tests.integration_tests.subroutines import run_inference_engines


class TestProblem(unittest.TestCase):

    def test_linear_regression(self, n_steps=200, n_initial_steps=100,
                               n_walkers=20, plot=False, show_progress=False,
                               run_scipy=True, run_emcee=True, run_torch=True):
        """
        Integration test for the problem described at the top of this file.

        Parameters
        ----------
        n_steps : int, optional
            Number of steps (samples) to run. Note that the default number is
            rather low just so the test does not take too long.
        n_initial_steps : int, optional
            Number of steps for initial (burn-in) sampling.
        n_walkers : int, optional
            Number of walkers used by the estimator.
        plot : bool, optional
            If True, the data and the posterior distributions are plotted. This
            is deactivated by default, so that the test does not stop until the
            generated plots are closed.
        show_progress : bool, optional
            If True, progress-bars will be shown, if available.
        run_scipy : bool, optional
            If True, the problem is solved with scipy (maximum likelihood est).
            Otherwise, no maximum likelihood estimate is derived.
        run_emcee : bool, optional
            If True, the problem is solved with the emcee solver. Otherwise,
            the emcee solver will not be used.
        run_torch : bool, optional
            If True, the problem is solved with the pyro/torch_ solver.
            Otherwise, the pyro/torch_ solver will not be used.
        """

        # ==================================================================== #
        #                          Set numeric values                          #
        # ==================================================================== #

        # 'true' value of a, and its normal prior parameters
        a_true = 2.5
        loc_a = 2.0
        scale_a = 1.0

        # 'true' value of b, and its normal prior parameters
        b_true = 1.7
        loc_b = 1.0
        scale_b = 1.0

        # 'true' value of noise sd, and its uniform prior parameters
        sigma_noise = 0.5
        low_sigma = 0.1
        high_sigma = 0.8

        # the number of generated experiment_names and seed for random numbers
        n_tests = 50
        seed = 1

        # ==================================================================== #
        #                       Define the Forward Model                       #
        # ==================================================================== #

        class LinearModel(ForwardModelBase):
            def response(self, inp):
                x = inp['x']
                m = inp['m']
                b = inp['b']
                response = {}
                for os in self.output_sensors:
                    response[os.name] = m * x + b
                return response

        # ==================================================================== #
        #                     Define the Inference Problem                     #
        # ==================================================================== #

        # initialize the inference problem with a useful name; note that the
        # name will only be stored as an attribute of the InferenceProblem and
        # is not important for the problem itself; can be useful when dealing
        # with multiple problems
        problem = InferenceProblem("Linear regression with normal noise")

        # add all parameters to the problem; the first argument states the
        # parameter's global name (here: 'a', 'b' and 'sigma'); the second
        # argument defines the parameter type (three options: 'model' for
        # parameter's of the forward model, 'prior' for prior parameters and
        # 'noise' for parameters of the noise model); the 'info'-argument is a
        # short description string used for logging, and the tex-argument gives
        # a tex-string of the parameter used for plotting; finally, the prior-
        # argument specifies the parameter's prior; note that this definition
        # of a prior will result in the initialization of constant parameters of
        # type 'prior' in the background
        problem.add_parameter('a', 'model',
                              tex="$a$",
                              info="Slope of the graph",
                              prior=('normal', {'loc': loc_a,
                                                'scale': scale_a}))
        problem.add_parameter('b', 'model',
                              info="Intersection of graph with y-axis",
                              tex='$b$',
                              prior=('normal', {'loc': loc_b,
                                                'scale': scale_b}))
        problem.add_parameter('sigma', 'noise',
                              tex=r"$\sigma$",
                              info="Std. dev, of 0-mean noise model",
                              prior=('uniform', {'low': low_sigma,
                                                 'high': high_sigma}))

        # add the forward model to the problem; note that the first positional
        # argument [{'a': 'm'}, 'b'] passed to LinearModel defines the forward
        # model's parameters by name via a list with elements structured like
        # {<global parameter name>: <local parameter name>}; a global name is a
        # name introduced by problem.add_parameter, while a local name is a name
        # used in the response-method of the forward model class (see the class
        # LinearModel above); note that the use of the local parameter name 'm'
        # for the global parameter 'a' is added here only to highlight the
        # possibility of this feature; it is not necessary at all here; whenever
        # forward model's parameter has a similar local and global name (which
        # should be the case most of the times), one doesn't have to use the
        # verbose notation  {<global parameter name>: <local parameter name>}
        # but can instead just write the parameter's (global=local) name, like
        # it is done with the forward model's parameter 'b' below
        isensor = Sensor("x")
        osensor = Sensor("y")
        linear_model = LinearModel([{'a': 'm'}, 'b'], [isensor], [osensor])
        problem.add_forward_model("LinearModel", linear_model)

        # add the noise model to the problem
        problem.add_noise_model(NormalNoiseModel(
            prms_def={'sigma': 'std'}, sensors=osensor))

        # ==================================================================== #
        #                Add test data to the Inference Problem                #
        # ==================================================================== #

        # data-generation; normal noise with constant variance around each point
        np.random.seed(seed)
        x_test = np.linspace(0.0, 1.0, n_tests)
        y_true = linear_model.response(
            {isensor.name: x_test, 'm': a_true, 'b': b_true})[osensor.name]
        y_test = np.random.normal(loc=y_true, scale=sigma_noise)

        # add the experimental data
        problem.add_experiment(f'TestSeries_1', fwd_model_name="LinearModel",
                               sensor_values={isensor.name: x_test,
                                              osensor.name: y_test})

        # give problem overview
        problem.info()

        # plot the true and noisy data
        if plot:
            plt.scatter(x_test, y_test, label='measured data',
                        s=10, c="red", zorder=10)
            plt.plot(x_test, y_true, label='true', c="black")
            plt.xlabel(isensor.name)
            plt.ylabel(osensor.name)
            plt.legend()
            plt.tight_layout()
            plt.draw()  # does not stop execution

        # ==================================================================== #
        #                Solve problem with inference engine(s)                #
        # ==================================================================== #

        # this routine is imported from another script because it it used by all
        # integration tests in the same way; ref_values are used for plotting
        true_values = {'a': a_true, 'b': b_true, 'sigma': sigma_noise}
        run_inference_engines(problem, true_values=true_values, n_steps=n_steps,
                              n_initial_steps=n_initial_steps,
                              n_walkers=n_walkers, plot=plot,
                              show_progress=show_progress,
                              run_scipy=run_scipy, run_emcee=run_emcee,
                              run_torch=run_torch)

if __name__ == "__main__":
    unittest.main()



# module imports
from probeye.definition import inference_problem
from probeye.definition import forward_model
from probeye.definition import noise_model
from probeye.definition import parameter
from probeye.definition import prior



"""
Linear regression example where a prior parameter is a latent parameter
--------------------------------------------------------------------------------
The model equation is y = a * x + b with a, b being the model parameters and the
noise model is a normal zero-mean distribution with the std. deviation to infer.
Additionally, the location parameter of a's prior is considered a latent
parameter.The problem is solved via sampling using emcee and pyro.
"""

# standard library imports
import unittest


# third party imports
import numpy as np
import matplotlib.pyplot as plt

# local imports (problem definition)
from probeye.definition.inference_problem import InferenceProblem
from probeye.definition.forward_model import ForwardModelBase
from probeye.definition.sensor import Sensor
from probeye.definition.noise_model import NormalNoiseModel

# local imports (testing related)
from tests.integration_tests.subroutines import run_inference_engines


class TestProblem(unittest.TestCase):

    def test_prior_calibration(self, n_steps=200, n_initial_steps=100,
                               n_walkers=20, plot=False, show_progress=False,
                               run_scipy=True, run_emcee=True, run_torch=True):
        """
        Integration test for the problem described at the top of this file.

        Parameters
        ----------
        n_steps : int, optional
            Number of steps (samples) to run. Note that the default number is
            rather low just so the test does not take too long.
        n_initial_steps : int, optional
            Number of steps for initial (burn-in) sampling.
        n_walkers : int, optional
            Number of walkers used by the estimator.
        plot : bool, optional
            If True, the data and the posterior distributions are plotted. This
            is deactivated by default, so that the test does not stop until the
            generated plots are closed.
        show_progress : bool, optional
            If True, progress-bars will be shown, if available.
        run_scipy : bool, optional
            If True, the problem is solved with scipy (maximum likelihood est).
            Otherwise, no maximum likelihood estimate is derived.
        run_emcee : bool, optional
            If True, the problem is solved with the emcee solver. Otherwise,
            the emcee solver will not be used.
        run_torch : bool, optional
            If True, the problem is solved with the pyro/torch_ solver.
            Otherwise, the pyro/torch_ solver will not be used.
        """

        # ==================================================================== #
        #                          Set numeric values                          #
        # ==================================================================== #

        # 'true' value of a, and its normal prior parameters
        a_true = 2.5
        scale_a = 1.0

        # uniform prior-parameters of 'loc_a'
        low_loc_a = 2.0
        high_loc_a = 3.0

        # 'true' value of b, and its normal prior parameters
        b_true = 1.7
        loc_b = 1.0
        scale_b = 1.0

        # 'true' value of noise sd, and its uniform prior parameters
        sigma_noise = 0.5
        low_sigma = 0.1
        high_sigma = 0.6

        # the number of generated experiment_names and seed for random numbers
        n_tests = 100
        seed = 1

        # ==================================================================== #
        #                       Define the Forward Model                       #
        # ==================================================================== #

        class LinearModel(ForwardModelBase):
            def response(self, inp):
                x = inp['x']
                a = inp['a']
                b = inp['b']
                response = {}
                for os in self.output_sensors:
                    response[os.name] = a * x + b
                return response

        # ==================================================================== #
        #                     Define the Inference Problem                     #
        # ==================================================================== #

        # initialize the inference problem with a useful name
        problem = InferenceProblem(
            "Linear model with normal noise and prior-prior")

        # add all parameters to the problem
        problem.add_parameter('loc_a', 'prior',
                              info="Location parameter of normal prior for 'a'",
                              tex=r"$\mu_a^\mathrm{prior}$",
                              prior=('uniform', {'low': low_loc_a,
                                                 'high': high_loc_a}))
        problem.add_parameter('a', 'model',
                              info="Slope of the graph",
                              tex="$a$",
                              prior=('normal', {'loc': 'loc_a',
                                                'scale': scale_a}))
        problem.add_parameter('b', 'model',
                              info="Intersection of graph with y-axis",
                              tex='$b$',
                              prior=('normal', {'loc': loc_b,
                                                'scale': scale_b}))
        problem.add_parameter('sigma', 'noise',
                              info="Std. dev, of 0-mean noise model",
                              tex=r"$\sigma$",
                              prior=('uniform', {'low': low_sigma,
                                                 'high': high_sigma}))

        # add the forward model to the problem
        isensor = Sensor("x")
        osensor = Sensor("y")
        linear_model = LinearModel(['a', 'b'], [isensor], [osensor])
        problem.add_forward_model("LinearModel", linear_model)

        # add the noise model to the problem
        problem.add_noise_model(NormalNoiseModel(
            prms_def={'sigma': 'std'}, sensors=osensor))

        # ==================================================================== #
        #                Add test data to the Inference Problem                #
        # ==================================================================== #

        # data-generation; normal noise with constant variance around each point
        np.random.seed(seed)
        x_test = np.linspace(0.0, 1.0, n_tests)
        y_true = linear_model(
            {isensor.name: x_test, 'a': a_true, 'b': b_true})[osensor.name]
        y_test = np.random.normal(loc=y_true, scale=sigma_noise)

        # add the experimental data
        problem.add_experiment(f'TestSeries_1', fwd_model_name="LinearModel",
                               sensor_values={isensor.name: x_test,
                                              osensor.name: y_test})

        # give problem overview
        problem.info()

        # plot the true and noisy data
        if plot:
            plt.scatter(x_test, y_test, label='measured data',
                        s=10, c="red", zorder=10)
            plt.plot(x_test, y_true, label='true', c="black")
            plt.xlabel(isensor.name)
            plt.ylabel(osensor.name)
            plt.legend()
            plt.tight_layout()
            plt.draw()  # does not stop execution

        # ==================================================================== #
        #                Solve problem with inference engine(s)                #
        # ==================================================================== #

        # this routine is imported from another script because it it used by all
        # integration tests in the same way
        true_values = {'loc_a': a_true, 'a': a_true, 'b': b_true,
                       'sigma': sigma_noise}
        run_inference_engines(problem, true_values=true_values, n_steps=n_steps,
                              n_initial_steps=n_initial_steps,
                              n_walkers=n_walkers, plot=plot,
                              show_progress=show_progress,
                              run_scipy=run_scipy, run_emcee=run_emcee,
                              run_torch=run_torch)

if __name__ == "__main__":
    unittest.main()



"""
Linear model in time and space with three different noise models
--------------------------------------------------------------------------------
The model equation is y = A * x + B * t + c with A, B, c being the model prms.
while x and t represent position and time respectively. From the three model
parameters A and B are latent ones while c is a constant. Measurements are made
at three different positions (x-values) each of which is associated with an own
zero-mean, uncorrelated normal noise model with the std. deviations to infer.
This results in five latent parameters (parameters to infer). The problem
is solved via sampling by means of emcee and pyro.
"""

# standard library imports
import unittest

# third party imports
import numpy as np

# local imports (problem definition)
from probeye.definition.inference_problem import InferenceProblem
from probeye.definition.forward_model import ForwardModelBase
from probeye.definition.sensor import Sensor
from probeye.definition.noise_model import NormalNoiseModel

# local imports (testing related)
from tests.integration_tests.subroutines import run_inference_engines


class TestProblem(unittest.TestCase):

    def test_multiple_sensors(self, n_steps=200, n_initial_steps=100,
                              n_walkers=20, plot=False, show_progress=False,
                              run_scipy=True, run_emcee=True, run_torch=True):
        """
        Integration test for the problem described at the top of this file.

        Parameters
        ----------
        n_steps : int, optional
            Number of steps (samples) to run. Note that the default number is
            rather low just so the test does not take too long.
        n_walkers : int, optional
            Number of walkers used by the estimator.
        plot : bool, optional
            If True, the data and the posterior distributions are plotted. This
            is deactivated by default, so that the test does not stop until the
            generated plots are closed.
        show_progress : bool, optional
            If True, progress-bars will be shown, if available.
        run_scipy : bool, optional
            If True, the problem is solved with scipy (maximum likelihood est).
            Otherwise, no maximum likelihood estimate is derived.
        run_emcee : bool, optional
            If True, the problem is solved with the emcee solver. Otherwise,
            the emcee solver will not be used.
        run_torch : bool, optional
            If True, the problem is solved with the pyro/torch_ solver.
            Otherwise, the pyro/torch_ solver will not be used.
        """

        # ==================================================================== #
        #                          Set numeric values                          #
        # ==================================================================== #

        # 'true' value of A, and its normal prior parameters
        A_true = 1.3
        loc_A = 1.0
        scale_A = 1.0

        # 'true' value of B, and its normal prior parameters
        B_true = -1.0
        loc_B = -2.0
        scale_B = 1.5

        # 'true' value of sd_S1, and its uniform prior parameters
        sd_S1_true = 0.2
        low_S1 = 0.1
        high_S1 = 0.7

        # 'true' value of sd_S2, and its uniform prior parameters
        sd_S2_true = 0.4
        low_S2 = 0.1
        high_S2 = 0.7

        # 'true' value of sd_S3, and its uniform prior parameters
        sd_S3_true = 0.6
        low_S3 = 0.1
        high_S3 = 0.7

        # define sensor positions
        pos_s1 = 0.2
        pos_s2 = 0.5
        pos_s3 = 1.0

        # define global constant; this constant is used here only to test if
        # there are any problems when using global constants
        c = 0.5

        # ==================================================================== #
        #                       Define the Forward Model                       #
        # ==================================================================== #

        class LinearModel(ForwardModelBase):
            def response(self, inp):
                t = inp['time']
                A = inp['A']
                B = inp['B']
                const = inp['const']
                response = dict()
                for os in self.output_sensors:
                    response[os.name] = A * os.x + B * t + const
                return response

        # ==================================================================== #
        #                     Define the Inference Problem                     #
        # ==================================================================== #

        # initialize the inference problem with a useful name
        problem = InferenceProblem("Linear model with three noise models")

        # add all parameters to the problem
        problem.add_parameter('A', 'model',
                              prior=('normal', {'loc': loc_A,
                                                'scale': scale_A}),
                              info="Slope of the graph",
                              tex="$A$")
        problem.add_parameter('B', 'model',
                              prior=('normal', {'loc': loc_B,
                                                'scale': scale_B}),
                              info="Intersection of graph with y-axis",
                              tex='$B$')
        problem.add_parameter('sigma_1', 'noise',
                              prior=('uniform', {'low': low_S1,
                                                 'high': high_S1}),
                              info="Std. dev. of zero-mean noise model for S1",
                              tex=r"$\sigma_1$")
        problem.add_parameter('sigma_2', 'noise',
                              prior=('uniform', {'low': low_S2,
                                                 'high': high_S2}),
                              info="Std. dev. of zero-mean noise model for S1",
                              tex=r"$\sigma_2$")
        problem.add_parameter('sigma_3', 'noise',
                              prior=('uniform', {'low': low_S3,
                                                 'high': high_S3}),
                              info="Std. dev. of zero-mean noise model for S1",
                              tex=r"$\sigma_3$")
        problem.add_parameter('c', 'model', const=c)

        # add the forward model to the problem
        isensor = Sensor("time")
        osensor1 = Sensor("y1", x=pos_s1)
        osensor2 = Sensor("y2", x=pos_s2)
        osensor3 = Sensor("y3", x=pos_s3)
        linear_model = LinearModel(['A', 'B', {'c': 'const'}], [isensor],
                                   [osensor1, osensor2, osensor3])
        problem.add_forward_model("LinearModel", linear_model)

        # add the noise models to the problem
        problem.add_noise_model(NormalNoiseModel(
            prms_def={'sigma_1': 'std'}, sensors=osensor1))
        problem.add_noise_model(NormalNoiseModel(
            prms_def={'sigma_2': 'std'}, sensors=osensor2))
        problem.add_noise_model(NormalNoiseModel(
            prms_def={'sigma_3': 'std'}, sensors=osensor3))

        # ==================================================================== #
        #                Add test data to the Inference Problem                #
        # ==================================================================== #

        # add the experimental data
        np.random.seed(1)
        sd_dict = {osensor1.name: sd_S1_true,
                   osensor2.name: sd_S2_true,
                   osensor3.name: sd_S3_true}

        def generate_data(n_time_steps, n=None):
            time_steps = np.linspace(0, 1, n_time_steps)
            inp = {'A': A_true, 'B': B_true, 'const': c, 'time': time_steps}
            sensors = linear_model(inp)
            for key, val in sensors.items():
                sensors[key] = val + np.random.normal(0.0, sd_dict[key],
                                                      size=n_time_steps)
            sensors[isensor.name] = time_steps
            problem.add_experiment(f'TestSeries_{n}',
                                   sensor_values=sensors,
                                   fwd_model_name='LinearModel')
        for n_exp, n_t in enumerate([101, 51]):
            generate_data(n_t, n=n_exp)

        # give problem overview
        problem.info()

        # ==================================================================== #
        #                Solve problem with inference engine(s)                #
        # ==================================================================== #

        # this routine is imported from another script because it it used by all
        # integration tests in the same way
        true_values = {'A': A_true, 'B': B_true, 'sigma_1': sd_S1_true,
                       'sigma_2': sd_S2_true, 'sigma_3': sd_S3_true}
        run_inference_engines(problem, true_values=true_values, n_steps=n_steps,
                              n_initial_steps=n_initial_steps,
                              n_walkers=n_walkers, plot=plot,
                              show_progress=show_progress,
                              run_scipy=run_scipy, run_emcee=run_emcee,
                              run_torch=run_torch)

if __name__ == "__main__":
    unittest.main()



# local imports (inference engines)
from probeye.inference.scipy_.solver import ScipySolver
from probeye.inference.emcee_.solver import EmceeSolver
from probeye.inference.torch_.solver import PyroSolver

# local imports (post-processing)
from probeye.postprocessing.sampling import create_pair_plot
from probeye.postprocessing.sampling import create_posterior_plot
from probeye.postprocessing.sampling import create_trace_plot


def run_inference_engines(problem, true_values=None, n_steps=1000,
                          n_initial_steps=100, n_walkers=20, plot=True,
                          show_progress=True, run_scipy=True, run_emcee=True,
                          run_torch=True):
    """
    Runs a requested selection of inference engines on a given problem. This
    function is in a separate file to avoid repeating the same lines of code in
    all the integration tests.

    Parameters
    ----------
    problem : obj[InferenceProblem]
        Describes the inference problem including e.g. parameters and data.
    true_values : None, dict, optional
        Used for plotting 'true' parameter values. Keys are the parameter names
        and values are the values that are supposed to be shown in the marginal
        plots.
    n_steps : int, optional
        Number of steps (samples) to run.
    n_initial_steps : int, optional
        Number of steps for initial (burn-in) sampling.
    n_walkers : int, optional
        Number of walkers used by the estimator.
    plot : bool, optional
        If True, the data and the post-processing plots are plotted.
    show_progress : bool, optional
        If True, progress-bars will be shown, if available.
    run_scipy : bool, optional
        If True, the problem is solved with scipy (maximum likelihood estimate).
        Otherwise, no maximum likelihood estimate is derived.
    run_emcee : bool, optional
        If True, the problem is solved with the emcee solver. Otherwise, the
        emcee solver will not be used.
    run_torch : bool, optional
        If True, the problem is solved with the pyro/torch_ solver. Otherwise,
        the pyro/torch_ solver will not be used.
    """

    # this loop avoids to write down the same lines of code for the post-
    # processing for each inference engine again
    for inference_engine, requested_to_run in {'scipy': run_scipy,
                                               'emcee': run_emcee,
                                               'pyro':  run_torch}.items():

        # solve the problem with scipy (max. likelihood) if requested
        if inference_engine == 'scipy':
            is_sampling_solver = False
            if requested_to_run:
                scipy_solver = ScipySolver(problem, show_progress=show_progress)
                inference_data = scipy_solver.run_max_likelihood(
                    true_values=true_values)
            else:
                # in this case, the engine was not requested to run
                continue

        # solve the problem with emcee if requested
        elif inference_engine == 'emcee':
            is_sampling_solver = True
            if requested_to_run:
                emcee_solver = EmceeSolver(problem, show_progress=show_progress)
                inference_data = emcee_solver.run_mcmc(
                    n_walkers=n_walkers, n_steps=n_steps,
                    n_initial_steps=n_initial_steps)
            else:
                # in this case, the engine was not requested to run
                continue

        # solver the problem with pyro/torch if requested
        elif inference_engine == 'pyro':
            is_sampling_solver = True
            if requested_to_run:
                n_walkers_used = 1  # getting errors when trying to use more
                pyro_solver = PyroSolver(problem, show_progress=show_progress)
                inference_data = pyro_solver.run_mcmc(
                    n_walkers=n_walkers_used, n_steps=n_steps,
                    n_initial_steps=n_initial_steps)
            else:
                # in this case, the engine was not requested to run
                continue

        else:
            # if this happens, you messed something up in the defining
            # dictionary of this loop
            raise RuntimeError(
                f"Found unknown inference engine '{inference_engine}'!")

        # do the post-processing; note that the interface is the same for
        # each inference engine; also note, that the plots are not so much
        # intended for automatic testing, as for manually running the script
        if plot:
            if is_sampling_solver:
                create_pair_plot(inference_data, problem,
                                 true_values=true_values)
                create_posterior_plot(inference_data, problem)
                create_trace_plot(inference_data, problem)



# ============================================================================ #
#                                   Imports                                    #
# ============================================================================ #

# standard library imports
from copy import copy
from typing import Iterable
import os
import sys

# third party imports
import numpy as np
import configparser
from loguru import logger
from functools import partial

logger = logger.opt(colors=True)
logger.opt = partial(logger.opt, colors=True)


# ============================================================================ #
#                                 Subroutines                                  #
# ============================================================================ #

def len_or_one(obj):
    """
    Returns the length of an object or 1 if no length is defined.

    Parameters
    ----------
    obj : obj
        Most of the time this will be a list/tuple or a single scalar number.

    Returns
    -------
    int
        The length of the given list/tuple etc. or 1, if obj has no __len__
        attribute; the latter case is mostly intended for scalar numbers

    """
    if hasattr(obj, '__len__'):
        # the following check is necessary, since the len-function applied to a
        # numpy array of format numpy.array(1) results in a TypeError
        if type(obj) is np.ndarray:
            if not obj.shape:
                return 1
            else:
                return len(obj)
        else:
            return len(obj)
    else:
        return 1

def make_list(arg):
    """
    Converts a given argument into a list, if it is not a list or tuple. The
    typical use case for this method is to convert a single string into a list
    with this string as its only element, e.g. make_list('sigma') = ['sigma'].

    Parameters
    ----------
    arg : obj
        Essentially anything. Most of the time this might be a string or a list
        of strings or a tuple of strings.

    Returns
    -------
    new_arg : list[obj]
        Either arg if it is of type list or tuple, or a list with arg as its
        only element if arg is not of type list or tuple.
    """
    if type(arg) in [list, tuple]:
        new_arg = arg
    else:
        new_arg = [copy(arg)]
    return new_arg

def underlined_string(string, symbol="=", n_empty_start=1, n_empty_end=1):
    """
    Adds a line made of 'symbol'-characters under a given string and returns it.

    Parameters
    ----------
    string : string
        The string that should be underlined.
    symbol : string
        A single character the line should be 'made' of.
    n_empty_start : int, optional
        Number of empty lines added before the underlined string.
    n_empty_end : int, optional
        Number of empty lines added after the underlined string.

    Returns
    -------
    result_string : string
        The generated string representing an underlined string, possibly with
        empty lines added before/after.
    """
    n_chars = len(string)
    underline_string = n_chars * symbol
    empty_lines_start = n_empty_start * "\n"
    empty_lines_end = n_empty_end * "\n"
    result_string = string + "\n" + underline_string
    result_string = empty_lines_start + result_string + empty_lines_end
    return result_string

def titled_table(title_str, table_str, symbol='-', n_empty_start=1,
                 n_empty_end=0):
    """
    Adds an underlined title string to a given table string. The line, that
    underlines the title will be as long as the longest line of the table.

    Parameters
    ----------
    title_str : string
        The title to be put on top of the table.
    table_str : string
        The string representing the table. For example generated using tabulate.
    symbol : string
        A single character the line should be 'made' of.
    n_empty_start : int, optional
        Number of empty lines added before the title.
    n_empty_end : int, optional
        Number of empty lines added after the table string.

    Returns
    -------
    result_string : string
        An underlined title, followed by a table.
    """
    # get the number of characters in the given table's longest line
    max_line_length = max([len(line) for line in table_str.split('\n')])
    # now, simply concatenate the different lines
    result_string = n_empty_start * '\n' +\
                    title_str + '\n' +\
                    max_line_length * symbol + '\n' +\
                    table_str + '\n' +\
                    n_empty_end * '\n'
    return result_string

def replace_string_chars(string, replace=None, remove=None):
    """
    Removes and replaces characters from a given string according to the input.

    Parameters
    ----------
    string : string
        The string to be modified.
    replace : dict, optional
        The keys are the characters to be replaced, the values are stating their
        replacements.
    remove : list, optional
        A list of characters to be removed from the given string.

    Returns
    -------
     string : string
        The modified string with removed/replaced characters.
    """
    # first, do the replacements
    if replace is not None:
        for char, replacement in replace.items():
            string = string.replace(char, replacement)
    # finally, remove characters as requested
    if remove is not None:
        for char in remove:
            string = string.replace(char, '')
    return string

def simplified_list_string(list_):
    """
    Given a list, it returns the string returned by its __str__ method, where
    some characters are removed for a slightly easier to read printout. Example:
    the list ['a', 1.2] is usually printed as '['a', 1.2]'. Here, it will be
    converted into the string 'a, 1.2' (no brackets, and no quotes).

    Parameters
    ----------
    list_ : list
        Some list to be printed.

    Returns
    -------
    simplified_list_str : string
        The list_'s __str__ method's return string without brackets and quotes.
    """
    simplified_list_str = replace_string_chars(
        str(list_), remove=['[', ']', "'"])
    return simplified_list_str

def simplified_dict_string(dict_):
    """
    Given a dictionary, it returns the string returned by its __str__ method,
    where some characters are removed for a slightly easier to read printout.
    Example: the dict {'a': 1.2} is usually printed as '{'a': 1.2, 'b': 2.1}'.
    Here, it will be converted into the string 'a=1.2, b=2.1'.

    Parameters
    ----------
    dict_ : dict
        Some dictionary to be printed.

    Returns
    -------
    simplified_dict_str : string
        Modified version of dict_'s __str__ method's return string (no quotes,
        no braces and the colon will be replaced with an equal sign.
    """
    simplified_dict_str = replace_string_chars(
        str(dict_), remove=['{', '}', "'"], replace={': ': '='})
    return simplified_dict_str

def unvectorize_dict_values(dict_):
    """
    Takes a dict with items like <name>: <vector> and converts it into a list,
    where each element is a 'fraction' or the whole dictionary. The following
    example will illustrate it: {'x': [1, 2, 3], 'y': [4, 5, 6]} will be
    converted into [{'x': 1, 'y': 4}, {'x': 2, 'y': 5}, {'x': 3, 'y': 6}].

    Parameters
    ----------
    dict_ : dict
        The dictionary that should be converted. All values must be 1D arrays
        of the same length.

    Returns
    -------
    result_list : list
        The 'un-vectorized' dictionary. Check out the example above.
    """

    # all values must be iterable
    dict_copy = copy(dict_)
    for key, value in dict_.items():
        if not hasattr(value, '__len__'):
            dict_copy[key] = [value]

    # check if all lengths are the same
    if len({len(vector) for vector in dict_copy.values()}) != 1:
        raise RuntimeError(
            "The values of the dictionary have different lengths!"
        )

    # create the result list
    vector_length = len([*dict_copy.values()][0])
    keys = [*dict_.keys()]
    result_list = []
    for i in range(vector_length):
        atom_dict = dict()
        for key in keys:
            atom_dict[key] = dict_copy[key][i]
        result_list.append(atom_dict)

    return result_list

def sub_when_empty(string, empty_str="-"):
    """
    Just returns a given string if it is not empty. If it is empty though, a
    default string is returned instead.

    Parameters
    ----------
    string : string
        The string to check if it is empty or not
    empty_str : string, optional
        The string to be returned if the given string is empty

    Returns
    -------
    result_string : string
        Either the given string (when 'string' is not empty) or the empty_str
        (when 'string' is empty)
    """
    if type(string) is not str:
        raise TypeError(
            f"Input must be of type string. Found: type '{type(string)}'")
    if len(string) > 0:
        result_string = string
    else:
        result_string = empty_str
    return result_string

def dict2list(dict_):
    """
    Converts a dict into a list of key-value dictionaries and returns it.

    Parameters
    ----------
    dict_ : dict
        Some dictionary to be converted.

    Returns
    -------
    list_ : list
        Each element is a dict with one key-value pair. These key-value pairs
        are those contained in dict_.
    """
    if type(dict_) != dict:
        raise TypeError(
            f"Input argument must be of type 'dict', found '{type(dict_)}'.")
    list_ = []
    for key, value in dict_.items():
        list_.append({key: value})
    return list_

def list2dict(list_dict):
    """
    Converts a list into a specific dictionary. The list may only contain
    strings or one-element dictionaries. For example [{'a': 'm'}, 'b'] will be
    converted into {'a': 'm', 'b': 'b'}.

    Parameters
    ----------
    list_dict : list or dict
        If it's a list it may only contain strings or one-element dictionaries.

    Returns
    -------
    dict_ : dict
        Strings are mapped to themselves, while one-element dictionaries are
        simply added to this result dictionary.
    """
    # check the given input
    if type(list_dict) not in [list, dict]:
        raise TypeError(
            f"The input argument must be of type 'list' or 'dict'. Found type "
            f"'{type(list_dict)}' however."
        )
    if type(list_dict) is dict:
        # convert the dict to a list, so it can be checked by this function
        list_ = dict2list(copy(list_dict))
    else:
        list_ = copy(list_dict)
    # convert the list to a dictionary
    dict_ = {}
    for element in list_:
        element_type = type(element)
        if element_type == dict:
            if len(element) != 1:
                raise ValueError(
                    f"Found a dict-element, which has {len(element)} instead "
                    f"of 1 key-value pair."
                )
            dict_ = {**dict_, **element}
        elif element_type == str:
            dict_[element] = element
        else:
            raise TypeError(
                f"The elements in the given list must be either of type "
                f"'string' or 'dict'. Found '{element_type}' however."
            )
    return dict_

def pretty_time_delta(seconds):
    """
    Converts number of seconds into a human friendly time string. Source: https:
    //gist.github.com/thatalextaylor/7408395#file-1-python-pretty-time-delta-py

    Parameters
    ----------
    seconds : float, int
        The given number of seconds to be converted,.

    Returns
    -------
    str
        Human friendly time difference in string format.
    """
    sign_string = "-" if seconds < 0 else ""
    seconds = abs(int(seconds))
    days, seconds = divmod(seconds, 86400)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    if days > 0:
        return "%s%dd%dh%dm" % (sign_string, days, hours, minutes)
    elif hours > 0:
        return "%s%dh%dm%ds" % (sign_string, hours, minutes, seconds)
    elif minutes > 0:
        return "%s%dm%ds" % (sign_string, minutes, seconds)
    else:
        return "%s%ds" % (sign_string, seconds)

def flatten_generator(items):
    """
    Yield items from any nested iterable. This solution is modified from a
    recipe in Beazley, D. and B. Jones. Recipe 4.14, Python Cookbook 3rd Ed.,
    O'Reilly Media Inc. Sebastopol, CA: 2013.

    Parameters
    ----------
    items : Iterable
        A list, tuple, numpy.ndarray, etc. that should be flattened.

    Returns
    -------
    obj[generator]
        Can be translated to a list by applying list(...) on it.
    """
    for x in items:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            for sub_x in flatten_generator(x):
                yield sub_x
        else:
            yield x

def flatten(arg):
    """
    Flattens and returns the given input argument.

    Parameters
    ----------
    arg : list, numpy.ndarray, float, int, None
        The list/array that should be flattened.

    Returns
    -------
    arg_flat : list, None
        The flattened list/numpy.ndarray is the input is not None. Otherwise,
        None is returned.
    """
    arg_type = type(arg)
    if arg is None:
        arg_flat = arg
    elif arg_type in [float, int]:
        arg_flat = [arg]
    else:
        if arg_type not in [list, np.ndarray]:
            raise TypeError(f"The argument must be either None or of type list "
                            f"numpy.ndarray, float or int. Found type "
                            f"{arg_type} however.")
        arg_flat = list(flatten_generator(arg))
    return arg_flat

def process_spatial_coordinates(x=None, y=None, z=None, coords=None,
                                order=('x', 'y', 'z')):
    """
    x : float, int, numpy.ndarray, None, optional
        Positional x-coordinate. When given, the coords-argument must be None.
    y : float, int, numpy.ndarray, None, optional
        Positional y-coordinate. When given, the coords-argument must be None.
    z : float, int, numpy.ndarray, None, optional
        Positional z-coordinate. When given, the coords-argument must be None.
    coords : numpy.ndarray, optional
        Some or all of the coordinates x, y, z concatenated as an array. Each
        row corresponds to one coordinate. For example, row 1 might contain all
        x-coordinates. Which row corresponds to which coordinate is defined via
        the order-argument. When the coords-argument is given, all 3 arguments
        x, y and z must be None.
    order : tuple[str], optional
        Only relevant when coords is given. Defines which row in coords
        corresponds to which coordinate. For example, order=('x', 'y', 'z')
        means that the 1st row are x-coordinates, the 2nd row are y-coords and
        the 3rd row are the z-coordinates.

    Returns
    -------
    coords : numpy.ndarray
        An array with as many columns as coordinates are given, and as many rows
        as points are given. For example if 10 points with x and z coordinates
        are given, then coords would have a shape of (2, 10).
    adjusted_order : list[str]
        Describes which coordinates are described by the rows of the returned
        coords. In the example given above, adjusted_order would be ['x', 'z'].
    """

    # the following check should cover the option that no spatial input is given
    if (x is None) and (y is None) and (z is None) and (coords is None):
        return np.array([]), []

    # convert all single-coordinate inputs to flat numpy arrays
    x = np.array(flatten(x)) if x is not None else None
    y = np.array(flatten(y)) if y is not None else None
    z = np.array(flatten(z)) if z is not None else None

    # derive the number of given coordinate vectors and points
    if coords is not None:
        if not type(coords) is np.ndarray:
            raise TypeError(f"The argument 'coords' must be of type "
                            f"numpy.ndarray. Found {type(coords)} however.")
        else:
            # each row corresponds to one coordinate, so the number of given
            # points is the length of rows
            n_coords, n_points = coords.shape
    else:
        n_points_list = [len(v) for v in [x, y, z] if v is not None]
        n_points_set = set(n_points_list)
        if len(n_points_set) == 1:
            n_coords = len(n_points_list)
            n_points = n_points_list[0]
        else:
            raise RuntimeError(
                f"Found inconsistent lengths in given coordinate "
                f"vectors: {n_points_list}!")

    # derive the coords array and the corresponding order-vector to be returned;
    # note that the repeated if-else clause here should improve readability
    if coords is not None:
        # it is assumed here that the first n_coords elements from the order-
        # vector correspond to the n_coords rows of the given coords-argument
        adjusted_order = list(order[:n_coords])
    else:
        # in this case the order-vector might have to be trimmed; for example if
        # x and z are given, the 'y' from the order vector has to be removed
        coords = np.zeros((n_coords, n_points))
        adjusted_order = []
        row_idx = 0
        for v in order:
            if eval(v) is not None:
                adjusted_order.append(v)
                coords[row_idx, :] = eval(v)
                row_idx += 1

    return coords, adjusted_order

def translate_prms_def(prms_def_given):
    """
    Translates the prms_def argument which is used by several sub-modules (e.g.
    ForwardModelBase, NoiseModelBase, PriorBase) into a default format. The
    prms_def argument specifies the local/global names of the parameters used by
    a sub-module.

    Parameters
    ----------
    prms_def_given : {str, list, dict}
        Either a single string, a dictionary with global names as keys and local
        names as values, or a list, the elements of which are either strings or
        1-element dictionaries, where the latter would again contain one global
        name as key and one local name as value. Valid examples are: 'sigma',
        ['sigma'], ['sigma', 'beta'], ['sigma', {'beta': 'b'}],
        {'sigma': 'sigma', 'beta': 'b'}.

    Returns
    -------
    prms_def : dict
        Contains global names as keys and local names as values.
    prms_dim : int
        The number of items in prms_def.
    """
    prms_def_copy = copy(prms_def_given)
    if type(prms_def_copy) is dict:
        prms_def = list2dict(prms_def_copy)
    else:
        prms_def = list2dict(make_list(prms_def_copy))
    prms_dim = len(prms_def)
    return prms_def, prms_dim

def print_probeye_header(width=100, header_file="../probeye.txt",
                         setup_cfg="../setup.cfg", margin=5, h_symbol="=",
                         v_symbol="#", use_logger=True):
    """
    Prints the probeye header which is printed, when an inference problem is
    set up. Mostly just nice to have. The only useful information it contains
    is the version number of the package.

    Parameters
    ----------
    width : int, optional
        The width (i.e., number of characters) the header should have.
    header_file : str, optional
        Relative path (with respect to this file) to the txt-file that contains
        the probeye letters.
    setup_cfg : str, optional
        Relative path (with respect to this file) to the setup.cfg file
        containing version number and description.
    margin : int, optional
        Minimum number of blank spaces at the header margins.
    h_symbol : str, optional
        The symbol used to 'draw' the horizontal frame line.
    v_symbol : str, optional
        The symbol used to 'draw' the vertical frame line.
    use_logger : bool, optional
        When True, the header will be logged, otherwise just printed.
    """

    # define the full paths of the given files
    file_path = os.path.dirname(__file__)
    header_file = os.path.join(file_path, header_file)
    setup_cfg = os.path.join(file_path, setup_cfg)

    # read in the big probeye letters
    with open(header_file, 'r') as f:
        content = f.readlines()
    # this is the width of the read in 'probeye' in terms of number of chars;
    # note that all lines (should) have the same length
    width_probeye = len(content[0]) - 1

    # get the version and the description from the setup.cfg file
    cfg = configparser.ConfigParser()
    cfg.read(setup_cfg)
    version = cfg.get('metadata', 'version')
    description = cfg.get('metadata', 'description')

    subtitle = f"Version {version} - {description}"
    width_subtitle = len(subtitle)

    # choose a width so that the margin on one side is at least 'margin'
    width_used = max((width, width_probeye + 2 * (margin + 1),
                      width_subtitle + 2 * margin + 1))

    # assemble the header
    outer_frame_line = f"{v_symbol} {h_symbol * (width_used - 4)} {v_symbol}"
    inner_frame_line = f"{v_symbol}{' ' * (width_used - 2)}{v_symbol}"
    lines = [outer_frame_line, inner_frame_line]
    for line in content:
        clean_line = line.replace('\n', '')
        lines.append(f"{v_symbol}{clean_line:^{width_used - 2}s}{v_symbol}")
    lines.append(inner_frame_line)
    lines.append(outer_frame_line)
    lines.append(inner_frame_line)
    lines.append(f"{v_symbol}{subtitle:^{width_used - 2}s}{v_symbol}")
    lines.append(inner_frame_line)
    lines.append(outer_frame_line)

    # log or print the header
    if use_logger:
        print("")
        for line in lines:
            logger.info(line)
    else:
        print('\n' + '\n'.join(lines))

def logging_setup(log_level_stdout='INFO', log_level_file='DEBUG',
                  log_format=None, log_file=None, overwrite_log_file=True,
                  **kwargs):
    """
    Sets up the loguru logger for listening to the inference problem.

    Parameters
    ----------
    log_level_stdout : {'DEBUG', 'INFO', 'WARNING', 'ERROR'}, optional
        Defines the level of the logging output to stdout.
    log_level_file : {'DEBUG', 'INFO', 'WARNING', 'ERROR'}, optional
        Defines the level of the logging output to a log file.
    log_format : None, str, optional
        A format string defining the logging output. If this argument is
        set to None, a default format will be used.
    log_file : None, str, optional
        Path to the log-file, if the logging should be printed to file. If
        None is given, no logging-file will be created.
    overwrite_log_file : bool, optional
        When True, a specified log-file will be overwritten. Otherwise,
        the generated logging will appended to a given log-file.
    kwargs : dict
        Additional keyword arguments passed to logger.add (for file and stdout).
    """
    if not log_format:
        log_format =\
            ('<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | '
             '<level>{level: <8}</level> | '
             '<level>{message:100s}</level> | '
             '<cyan>{name}</cyan>:'
             '<cyan>{function}</cyan>:'
             '<cyan>{line}</cyan>')
    logger.remove()  # just in case there still exists another logger
    logger.add(sys.stdout, format=log_format, level=log_level_stdout, **kwargs)
    if log_file:
        if os.path.isfile(log_file) and overwrite_log_file:
            os.remove(log_file)
        logger.add(
            log_file, format=log_format, level=log_level_file, **kwargs)

def stream_to_logger(log_level):
    """
    Returns a stream-object that can be used to redirect a function's print
    output to the logger. Taken from the section 'Capturing standard stdout ...'
    of https://loguru.readthedocs.io/en/stable/resources/recipes.html.

    Parameters
    ----------
    log_level : {'DEBUG', 'INFO', 'WARNING', 'ERROR'}, optional
        Defines the log level the streamed output will be associated with.

    Returns
    -------
    obj[StreamToLogger]
        This object should be used as follows:
        import contextlib
        with contextlib.redirect_stdout(stream_to_logger('INFO')):
            <function that prints something>
    """

    class StreamToLogger:
        def __init__(self, level):
            self._level = level

        def write(self, buffer):
            for line in buffer.rstrip().splitlines():
                logger.opt(depth=1).log(self._level, line.rstrip())

        def flush(self):
            pass

    return StreamToLogger(log_level)

def print_dict_in_rows(d, printer=print, sep="=", val_fmt=None):
    """
    Prints a dictionary with key-value pairs in rows.

    Parameters
    ----------
    d : dict
        The dictionary to print.
    printer : callable, optional
        Function used for printing. For example 'print' or 'logger.info'.
    sep : str, optional
        The character printed between key and value.
    val_fmt : str, optional
        A format string used for printing the dictionary's values.
    """
    n = max([len(key) for key in d.keys()])
    for key, val in d.items():
        if val_fmt:
            printer(f"{key:{n + 1}s} {sep} {val:{val_fmt}}")
        else:
            printer(f"{key:{n + 1}s} {sep} {val}")



"""
Linear regression example with spatial correlation model
--------------------------------------------------------------------------------
The n data points (y1, y2, ..., yn) generated for this example are sampled from
an n-variate normal distribution with mean values given by yi = a * xi + b with
a, b being the model parameters and x1, x2, ..., xi, ..., xn being predefined
spatial x-coordinates ranging from 0 to 1. The data points (y1, y2, ..., yn) are
not independent but correlated. The corresponding covariance matrix is defined
based on an exponential correlation function parameterized by the const standard
deviation sigma of the n-variate normal distribution and a correlation length
l_corr. Hence, the full model has four parameters a, b, sigma, l_corr, all of
which are inferred in this example using emcee-sampling.
"""

# standard library
import unittest

# third party imports
import numpy as np
import matplotlib.pyplot as plt

# local imports (problem definition)
from probeye.definition.inference_problem import InferenceProblem
from probeye.definition.forward_model import ForwardModelBase
from probeye.definition.sensor import Sensor
from probeye.definition.noise_model import NormalNoiseModel

# local imports (testing related)
from tests.integration_tests.subroutines import run_inference_engines
from probeye.inference.scipy_.correlation_models import \
    SpatialExponentialCorrelationModel


class TestProblem(unittest.TestCase):

    def test_spatial_correlation(self, n_steps=200, n_initial_steps=100,
                                 n_walkers=20, plot=False, show_progress=False,
                                 run_scipy=True, run_emcee=True,
                                 run_torch=False):
        """
        Integration test for the problem described at the top of this file.

        Parameters
        ----------
        n_steps : int, optional
            Number of steps (samples) to run. Note that the default number is
            rather low just so the test does not take too long.
        n_initial_steps : int, optional
            Number of steps for initial (burn-in) sampling.
        n_walkers : int, optional
            Number of walkers used by the estimator.
        plot : bool, optional
            If True, the data and the posterior distributions are plotted. This
            is deactivated by default, so that the test does not stop until the
            generated plots are closed.
        show_progress : bool, optional
            If True, progress-bars will be shown, if available.
        run_scipy : bool, optional
            If True, the problem is solved with scipy (maximum likelihood est).
            Otherwise, no maximum likelihood estimate is derived.
        run_emcee : bool, optional
            If True, the problem is solved with the emcee solver. Otherwise,
            the emcee solver will not be used.
        run_torch : bool, optional
            If True, the problem is solved with the pyro/torch_ solver.
            Otherwise, the pyro/torch_ solver will not be used.
        """

        if run_torch:
            raise RuntimeError("The pyro-solver is not available yet for "
                               "forward models including correlations.")

        # ==================================================================== #
        #                          Set numeric values                          #
        # ==================================================================== #

        # 'true' value of a, and its normal prior parameters
        a_true = 2.5
        loc_a = 2.0
        scale_a = 1.0

        # 'true' value of b, and its normal prior parameters
        b_true = 1.7
        loc_b = 1.0
        scale_b = 1.0

        # 'true' value of noise sd, and its uniform prior parameters
        sigma = 0.5
        low_sigma = 0.1
        high_sigma = 0.8

        # 'true' value of correlation length, and its uniform prior parameters
        l_corr = 0.05
        low_l_corr = 0.001
        high_l_corr = 0.2

        # settings for the data generation
        n_experiments = 8
        n_points = 50
        seed = 1

        # ==================================================================== #
        #                       Define the Forward Model                       #
        # ==================================================================== #

        class LinearModel(ForwardModelBase):
            def response(self, inp):
                a = inp['a']
                b = inp['b']
                response = {}
                for os in self.output_sensors:
                    response[os.name] = a * os.x.flatten() + b
                return response

        # ==================================================================== #
        #                     Define the Inference Problem                     #
        # ==================================================================== #

        # initialize the inference problem with a useful name; note that the
        # name will only be stored as an attribute of the InferenceProblem and
        # is not important for the problem itself; can be useful when dealing
        # with multiple problems
        problem = InferenceProblem("Linear regression with normal noise")

        # add all parameters to the problem; the first argument states the
        # parameter's global name (here: 'a', 'b' and 'sigma'); the second
        # argument defines the parameter type (three options: 'model' for
        # parameter's of the forward model, 'prior' for prior parameters and
        # 'noise' for parameters of the noise model); the 'info'-argument is a
        # short description string used for logging, and the tex-argument gives
        # a tex-string of the parameter used for plotting; finally, the prior-
        # argument specifies the parameter's prior; note that this definition
        # of a prior will result in the initialization of constant parameters of
        # type 'prior' in the background
        problem.add_parameter('a', 'model',
                              tex="$a$",
                              info="Slope of the graph",
                              prior=('normal', {'loc': loc_a,
                                                'scale': scale_a}))
        problem.add_parameter('b', 'model',
                              info="Intersection of graph with y-axis",
                              tex='$b$',
                              prior=('normal', {'loc': loc_b,
                                                'scale': scale_b}))
        problem.add_parameter('sigma', 'noise',
                              tex=r"$\sigma$",
                              info="Std. dev, of 0-mean noise model",
                              prior=('uniform', {'low': low_sigma,
                                                 'high': high_sigma}))
        problem.add_parameter('l_corr', 'noise',
                              tex=r"$l_\mathrm{corr}$",
                              info="Correlation length of correlation model",
                              prior=('uniform', {'low': low_l_corr,
                                                 'high': high_l_corr}))

        # add the forward model to the problem; note that the first positional
        # argument [{'a': 'm'}, 'b'] passed to LinearModel defines the forward
        # model's parameters by name via a list with elements structured like
        # {<global parameter name>: <local parameter name>}; a global name is a
        # name introduced by problem.add_parameter, while a local name is a name
        # used in the response-method of the forward model class (see the class
        # LinearModel above); note that the use of the local parameter name 'm'
        # for the global parameter 'a' is added here only to highlight the
        # possibility of this feature; it is not necessary at all here; whenever
        # forward model's parameter has a similar local and global name (which
        # should be the case most of the times), one doesn't have to use the
        # verbose notation  {<global parameter name>: <local parameter name>}
        # but can instead just write the parameter's (global=local) name, like
        # it is done with the forward model's parameter 'b' below
        x_test = np.linspace(0.0, 1.0, n_points)
        osensor = Sensor("y", x=x_test)
        linear_model = LinearModel(['a', 'b'], [], [osensor])
        problem.add_forward_model("LinearModel", linear_model)

        # add the noise model to the problem
        noise_model = NormalNoiseModel(
            sensors=osensor, corr='x', corr_model='exp',
            prms_def=[{'sigma': 'std'}, 'l_corr'])
        problem.add_noise_model(noise_model)

        # ==================================================================== #
        #                Add test data to the Inference Problem                #
        # ==================================================================== #

        # data-generation; first create the true values without noise; these
        # true values will be the mean values for sampling from a multivariate
        # normal distribution
        np.random.seed(seed)
        y_true = linear_model({'a': a_true, 'b': b_true})[osensor.name]

        # create the covariance matrix
        correlation_model = SpatialExponentialCorrelationModel(x=osensor.x)
        cov = correlation_model({'std': sigma, 'l_corr': l_corr})

        # now generate the noisy test data including correlations; we assume
        # here that there are n_experiments test series
        for i in range(n_experiments):
            y_test = np.random.multivariate_normal(mean=y_true, cov=cov)
            problem.add_experiment(f'Test_{i}', fwd_model_name="LinearModel",
                                   sensor_values={osensor.name: y_test})
            if plot:
                plt.scatter(x_test, y_test, label=f'measured data (test {i+1})',
                            s=10, zorder=10)
        # finish the plot
        if plot:
            plt.plot(x_test, y_true, label='true model', c="black", linewidth=3)
            plt.xlabel('x')
            plt.ylabel(osensor.name)
            plt.legend()
            plt.tight_layout()
            plt.draw()  # plt.draw() does not stop execution

        # ==================================================================== #
        #                Solve problem with inference engine(s)                #
        # ==================================================================== #

        # this routine is imported from another script because it it used by all
        # integration tests in the same way
        true_values = {'a': a_true, 'b': b_true, 'sigma': sigma,
                       'l_corr': l_corr}
        run_inference_engines(problem, true_values=true_values, n_steps=n_steps,
                              n_initial_steps=n_initial_steps,
                              n_walkers=n_walkers, plot=plot,
                              show_progress=show_progress,
                              run_scipy=run_scipy, run_emcee=run_emcee,
                              run_torch=run_torch)

if __name__ == "__main__":
    unittest.main()



# standard library imports
import unittest

# third party imports
import numpy as np

# local imports
from probeye.definition.noise_model import NoiseModelBase
from probeye.definition.noise_model import NormalNoiseModel
from probeye.definition.sensor import Sensor


class TestProblem(unittest.TestCase):

    def test_init(self):
        # initialize an instance and check if everything is there
        noise_template = NoiseModelBase(
            'normal', ['bias', 'sigma'], Sensor('y'))
        self.assertEqual(noise_template.prms_def,
                         {'bias': 'bias', 'sigma': 'sigma'})
        self.assertEqual(noise_template.prms_dim, 2)
        noise_template = NoiseModelBase(
            'normal', ['b', {'sd': 'sigma'}], Sensor('y'))
        self.assertEqual(noise_template.prms_def,
                         {'b': 'b', 'sd': 'sigma'})
        self.assertEqual(noise_template.prms_dim, 2)
        self.assertEqual(noise_template.sensor_names, ['y'])
        self.assertEqual(noise_template.experiment_names, [])

    def test_add_experiment_names(self):
        # prepare the setup for the tests
        noise_template = NoiseModelBase(
            'normal', ['bias', 'sigma'], Sensor('y'))
        problem_experiments = {'Exp1': {'sensor_values': {'x': 1, 'y': 2},
                                        'forward_model': 'TestModel'},
                               'Exp2': {'sensor_values': {'x': 3, 'y': 4},
                                        'forward_model': 'TestModel'},
                               'Exp3': {'sensor_values': {'x': 5, 'y': 6},
                                        'forward_model': 'TestModel'},
                               'Exp4': {'sensor_values': {'x': 7, 'z': 8},
                                        'forward_model': 'TestModel'},
                               'Exp5': {'sensor_values': {'x': 7, 'y': 8},
                                        'forward_model': 'TestModel'},
                               'Exp6': {'sensor_values': {'x': 7, 'y': 8},
                                        'forward_model': 'OtherModel'}}
        # the following is usually done automatically when adding the noise
        # model to the inference problem
        noise_template.problem_experiments = problem_experiments
        # perform some simple tests for valid usage
        noise_template.add_experiments('Exp1')
        self.assertEqual(noise_template.experiment_names, ['Exp1'])
        noise_template.add_experiments(['Exp2', 'Exp3'])
        self.assertEqual(noise_template.experiment_names,
                         ['Exp1', 'Exp2', 'Exp3'])
        # now check some invalid input arguments
        with self.assertRaises(RuntimeError):
            # add an experiment, that does not have the noise model's sensors
            noise_template.add_experiments('Exp4')
        with self.assertRaises(RuntimeError):
            # add experiments that refer to more than one forward model
            noise_template.add_experiments(['Exp5', 'Exp6'])
        with self.assertRaises(RuntimeError):
            # adding the same experiment again
            noise_template.add_experiments('Exp1')

    def test_error(self):
        # prepare the setup for the tests
        noise_template = NoiseModelBase('normal', ['bias', 'sigma'],
                                        [Sensor('y1'), Sensor('y2')])
        # in the NoiseModelBase class the error_function is just a dummy,
        # so we need to define it to be able to test the error-method
        noise_template.error_function =\
            lambda d1, d2: {name: d1[name] - d2[name]
                            for name in noise_template.sensor_names}
        model_response_dict = {'Exp1': {'x': 1, 'y1': 3, 'y2': -3},
                               'Exp2': {'x': 3, 'y1': 6, 'y2': -6},
                               'Exp3': {'x': 5, 'y1': 9, 'y2': -9}}
        problem_experiments = {'Exp1': {'sensor_values': {'x': 1,
                                                          'y1': 2,
                                                          'y2': -2},
                                        'forward_model': 'TestModel'},
                               'Exp2': {'sensor_values': {'x': 3,
                                                          'y1': 4,
                                                          'y2': -4},
                                        'forward_model': 'TestModel'},
                               'Exp3': {'sensor_values': {'x': 5,
                                                          'y1': 6,
                                                          'y2': -6},
                                        'forward_model': 'TestModel'}}
        # the following is usually done automatically when adding the noise
        # model to the inference problem
        noise_template.problem_experiments = problem_experiments
        noise_template.add_experiments(['Exp1', 'Exp2', 'Exp3'])
        # now we can call the error-method
        computed_value = noise_template.error(model_response_dict)
        expected_value = {'y1': np.array([1., 2., 3.]),
                          'y2': np.array([-1., -2., -3.])}
        # due to the numpy arrays we have to loop over everything to compare
        for sensor_name, array in computed_value.items():
            for i in range(len(array)):
                self.assertEqual(array[i], expected_value[sensor_name][i])

    def test_normal_noise_init(self):
        # check valid initialization
        noise_model = NormalNoiseModel({'sigma': 'std'}, Sensor('y'))
        self.assertTrue(noise_model.zero_mean)
        self.assertEqual(noise_model.noise_type, 'additive')
        noise_model = NormalNoiseModel({'sigma': 'std'}, Sensor('y'),
                                       noise_type='multiplicative')
        self.assertEqual(noise_model.noise_type, 'multiplicative')
        # check invalid initialization
        with self.assertRaises(ValueError):
            # invalid noise_type given
            NormalNoiseModel(
                {'sigma': 'std'}, Sensor('y'), noise_type='invalid_type')

    def test_normal_noise_error_function(self):
        # simply check that the error_function is set correctly when the noise
        # model is initialized
        ym_dict, ye_dict = {'y': 4}, {'y': 2}
        noise_model = NormalNoiseModel({'sigma': 'std'}, Sensor('y'),
                                       noise_type='additive')
        computed_value = noise_model.error_function(ym_dict, ye_dict)
        expected_value = {'y': 2}
        self.assertEqual(computed_value, expected_value)
        noise_model = NormalNoiseModel({'sigma': 'std'}, Sensor('y'),
                                       noise_type='multiplicative')
        computed_value = noise_model.error_function(ym_dict, ye_dict)
        expected_value = {'y': 1.0}
        self.assertEqual(computed_value, expected_value)

    def test_normal_noise_error_function_additive(self):
        # check method for scalar values
        ym_dict, ye_dict = {'y': 4}, {'y': 2}
        noise_model = NormalNoiseModel({'sigma': 'std'}, Sensor('y'))
        computed_value = noise_model.error_function_additive(ym_dict, ye_dict)
        expected_value = {'y': 2.0}
        self.assertEqual(computed_value, expected_value)
        # check method for numpy-arrays
        ym_dict, ye_dict = {'y': np.array([4, 5])}, {'y': np.array([2, 4])}
        noise_model = NormalNoiseModel({'sigma': 'std'}, Sensor('y'))
        computed_value = noise_model.error_function_additive(ym_dict, ye_dict)
        expected_value = {'y': np.array([2, 1])}
        self.assertTrue(np.allclose(computed_value['y'], expected_value['y']))

    def test_normal_noise_error_function_multiplicative(self):
        # check method for scalar values
        ym_dict, ye_dict = {'y': 4}, {'y': 2}
        noise_model = NormalNoiseModel({'sigma': 'std'}, Sensor('y'))
        comp_value = noise_model.error_function_multiplicative(ym_dict, ye_dict)
        expected_value = {'y': 1.0}
        self.assertEqual(comp_value, expected_value)
        # check method for numpy-arrays
        ym_dict, ye_dict = {'y': np.array([4, 5])}, {'y': np.array([2, 4])}
        noise_model = NormalNoiseModel({'sigma': 'std'}, Sensor('y'))
        comp_value = noise_model.error_function_multiplicative(ym_dict, ye_dict)
        expected_value = {'y': np.array([1.0, 0.25])}
        self.assertTrue(np.allclose(comp_value['y'], expected_value['y']))

    # def test_normal_noise_loglike_contribution(self):
    #     # prepare the first setup for the tests (all add. errors are zero)
    #     noise_model = NormalNoiseModel({'sigma': 'std'}, 'y')
    #     model_response_dict = {'Exp1': {'x': 1, 'y': 3},
    #                            'Exp2': {'x': 3, 'y': 6},
    #                            'Exp3': {'x': 5, 'y': 9}}
    #     problem_experiments = {'Exp1': {'sensor_values': {'x': 1, 'y': 3},
    #                                     'forward_model': 'TestModel'},
    #                            'Exp2': {'sensor_values': {'x': 3, 'y': 6},
    #                                     'forward_model': 'TestModel'},
    #                            'Exp3': {'sensor_values': {'x': 5, 'y': 9},
    #                                     'forward_model': 'TestModel'}}
    #     # the following is usually done automatically when adding the noise
    #     # model to the inference problem
    #     noise_model.problem_experiments = problem_experiments
    #     noise_model.add_experiments(['Exp1', 'Exp2', 'Exp3'])
    #     sigma = 2.1
    #     prms = {'sigma': sigma}
    #     computed_result = noise_model.loglike_contribution(
    #         model_response_dict, prms)
    #     expected_result = 3 * (-log(sigma) - 1 / 2 * log(2 * pi))
    #     self.assertAlmostEqual(computed_result, expected_result)
    #
    #     # prepare the second setup for the tests (add. errors are non-zero)
    #     noise_model = NormalNoiseModel({'sigma': 'std'}, 'y')
    #     model_response_dict = {'Exp1': {'x': 1, 'y': 3},
    #                            'Exp2': {'x': 3, 'y': 6},
    #                            'Exp3': {'x': 5, 'y': 9}}
    #     problem_experiments = {'Exp1': {'sensor_values': {'x': 1, 'y': 4},
    #                                     'forward_model': 'TestModel'},
    #                            'Exp2': {'sensor_values': {'x': 3, 'y': 8},
    #                                     'forward_model': 'TestModel'},
    #                            'Exp3': {'sensor_values': {'x': 5, 'y': 7},
    #                                     'forward_model': 'TestModel'}}
    #     # the following is usually done automatically when adding the noise
    #     # model to the inference problem
    #     noise_model.problem_experiments = problem_experiments
    #     noise_model.add_experiments(['Exp1', 'Exp2', 'Exp3'])
    #     sigma = 1.1
    #     prms = {'sigma': sigma}
    #     computed_result = noise_model.loglike_contribution(
    #         model_response_dict, prms)
    #     expected_error = {'y': np.array([-1, -2, 2])}
    #     expected_result = 3 * (-log(sigma) - 1 / 2 * log(2 * pi)) - 1 / 2 \
    #                       * np.sum(np.square(expected_error['y'] / sigma))
    #     self.assertAlmostEqual(computed_result, expected_result)

if __name__ == "__main__":
    unittest.main()



"""
Inference problem with two forward models that share a common parameter
--------------------------------------------------------------------------------
The first model equation is y = a * x + b with a, b being the model parameters
and the second model equation is y = alpha * x**2 + b where alpha is a new model
parameter, and b is the same model parameter as in the first model equation.
Both forward models have the same noise model with a normal zero-mean
distribution where the standard deviation is to be inferred.The problem is
solved via sampling using emcee and pyro.
"""

# standard library imports
import unittest

# third party imports
import numpy as np
import matplotlib.pyplot as plt

# local imports (problem definition)
from probeye.definition.inference_problem import InferenceProblem
from probeye.definition.forward_model import ForwardModelBase
from probeye.definition.sensor import Sensor
from probeye.definition.noise_model import NormalNoiseModel

# local imports (testing related)
from tests.integration_tests.subroutines import run_inference_engines


class TestProblem(unittest.TestCase):

    def test_two_models(self, n_steps=200, n_initial_steps=100, n_walkers=20,
                        plot=False, show_progress=False, run_scipy=True,
                        run_emcee=True, run_torch=False):
        """
        Integration test for the problem described at the top of this file.

        Parameters
        ----------
        n_steps : int, optional
            Number of steps (samples) to run. Note that the default number is
            rather low just so the test does not take too long.
        n_initial_steps : int, optional
            Number of steps for initial (burn-in) sampling.
        n_walkers : int, optional
            Number of walkers used by the estimator.
        plot : bool, optional
            If True, the data and the posterior distributions are plotted. This
            is deactivated by default, so that the test does not stop until the
            generated plots are closed.
        show_progress : bool, optional
            If True, progress-bars will be shown, if available.
        run_scipy : bool, optional
            If True, the problem is solved with scipy (maximum likelihood est).
            Otherwise, no maximum likelihood estimate is derived.
        run_emcee : bool, optional
            If True, the problem is solved with the emcee solver. Otherwise,
            the emcee solver will not be used.
        run_torch : bool, optional
            If True, the problem is solved with the pyro/torch_ solver.
            Otherwise, the pyro/torch_ solver will not be used.
        """

        # ==================================================================== #
        #                          Set numeric values                          #
        # ==================================================================== #

        # 'true' value of a, and its normal prior parameters
        a_true = 2.5
        loc_a = 2.0
        scale_a = 1.0

        # 'true' value of b, and its normal prior parameters
        b_true = 1.7
        loc_b = 1.0
        scale_b = 1.0

        # 'true' value of alpha, and its normal prior parameters
        alpha_true = 0.7
        loc_alpha = 2.0
        scale_alpha = 1.0

        # 'true' value of sigma, and its normal prior parameters
        sigma_true = 0.15
        low_sigma = 0.1
        high_sigma = 2.0

        # the number of generated experiment_names and seed for random numbers
        n_tests = 100
        seed = 1

        # ==================================================================== #
        #                      Define the Forward Models                       #
        # ==================================================================== #

        class LinearModel(ForwardModelBase):
            def response(self, inp):
                x = inp['x']
                a = inp['a']
                b = inp['b']
                response = {}
                for os in self.output_sensors:
                    response[os.name] = a * x + b
                return response

        class QuadraticModel(ForwardModelBase):
            def response(self, inp):
                x = inp['x']
                alpha = inp['alpha']
                beta = inp['beta']
                response = {}
                for os in self.output_sensors:
                    response[os.name] = alpha * x ** 2 + beta
                return response

        # ==================================================================== #
        #                     Define the Inference Problem                     #
        # ==================================================================== #

        # initialize the inference problem with a useful name
        problem = InferenceProblem(
            "Two models with shared parameter and normal noise")

        # add all parameters to the problem
        problem.add_parameter('a', 'model',
                              info="Slope of the graph in linear model",
                              tex='$a$ (linear)',
                              prior=('normal', {'loc': loc_a,
                                                'scale': scale_a}))
        problem.add_parameter('alpha', 'model',
                              info="Factor of quadratic term",
                              tex=r'$\alpha$ (quad.)',
                              prior=('normal', {'loc': loc_alpha,
                                                'scale': scale_alpha}))
        problem.add_parameter('b', 'model',
                              info="Intersection of graph with y-axis",
                              tex='$b$ (shared)',
                              prior=('normal', {'loc': loc_b,
                                                'scale': scale_b}))
        problem.add_parameter('sigma', 'noise',
                              tex=r"$\sigma$ (noise)",
                              info="Std. deviation of zero-mean noise model",
                              prior=('uniform', {'low': low_sigma,
                                                 'high': high_sigma}))

        # add the forward model to the problem
        isensor = Sensor("x")
        osensor_linear = Sensor("y_linear")
        osensor_quadratic = Sensor("y_quadratic")
        linear_model = LinearModel(['a', 'b'], [isensor], [osensor_linear])
        problem.add_forward_model("LinearModel", linear_model)
        quadratic_model = QuadraticModel(['alpha', {'b': 'beta'}],
                                         [isensor], [osensor_quadratic])
        problem.add_forward_model("QuadraticModel", quadratic_model)

        # add the noise model to the problem
        problem.add_noise_model(NormalNoiseModel(
            prms_def={'sigma': 'std'}, sensors=osensor_linear))
        problem.add_noise_model(NormalNoiseModel(
            prms_def={'sigma': 'std'}, sensors=osensor_quadratic))

        # ==================================================================== #
        #                Add test data to the Inference Problem                #
        # ==================================================================== #

        # data-generation; normal noise with constant variance around each point
        np.random.seed(seed)
        x_test = np.linspace(0.0, 1.0, n_tests)
        y_linear_true = linear_model({isensor.name: x_test,
                                      'a': a_true,
                                      'b': b_true})[osensor_linear.name]
        y_test_linear = np.random.normal(loc=y_linear_true, scale=sigma_true)
        y_quadratic_true = quadratic_model(
            {isensor.name: x_test,
             'alpha': alpha_true,
             'beta': b_true})[osensor_quadratic.name]
        y_test_quadratic = np.random.normal(loc=y_quadratic_true,
                                            scale=sigma_true)

        # add the experimental data
        problem.add_experiment(
            f'TestSeries_linear',
            sensor_values={isensor.name: x_test,
                           osensor_linear.name: y_test_linear},
            fwd_model_name="LinearModel")
        problem.add_experiment(
            f'TestSeries_quadratic',
            sensor_values={isensor.name: x_test,
                           osensor_quadratic.name: y_test_quadratic},
            fwd_model_name="QuadraticModel")

        # give problem overview
        problem.info()

        # plot the true and noisy data
        if plot:
            plt.scatter(x_test, y_test_linear, label='measured data (linear)',
                        s=10, c="red", zorder=10)
            plt.plot(x_test, y_linear_true, label='true (linear)', c="black")
            plt.scatter(x_test, y_test_quadratic, s=10, c="orange", zorder=10,
                        label='measured data (quadratic)')
            plt.plot(x_test, y_quadratic_true, label='true (quadratic)',
                     c="blue")
            plt.xlabel(isensor.name)
            plt.ylabel(f"{osensor_linear.name}, {osensor_quadratic.name}")
            plt.legend()
            plt.tight_layout()
            plt.draw()  # does not stop execution

        # ==================================================================== #
        #                Solve problem with inference engine(s)                #
        # ==================================================================== #

        # this routine is imported from another script because it it used by all
        # integration tests in the same way
        true_values = {'a': a_true, 'alpha': alpha_true, 'b': b_true,
                       'sigma': sigma_true}
        run_inference_engines(problem, true_values=true_values, n_steps=n_steps,
                              n_initial_steps=n_initial_steps,
                              n_walkers=n_walkers, plot=plot,
                              show_progress=show_progress,
                              run_scipy=run_scipy, run_emcee=run_emcee,
                              run_torch=run_torch)

if __name__ == "__main__":
    unittest.main()



# local imports
from probeye.subroutines import process_spatial_coordinates


class Sensor:
    """
    Base class for a sensor of the forward model. In its simplest form just a
    class with a 'name' attribute. Additional attributes for the measured
    quality (measurand) and the corresponding unit can be defined. If the
    sensors position(s) are important, they can be defined as attributes.
    Further attributes can be defined by the user by creating new classes
    derived from this one.
    """

    def __init__(self, name, measurand='not defined', unit='not defined',
                 x=None, y=None, z=None, coords=None, order=('x', 'y', 'z')):
        """
        Parameters
        ----------
        name : str
            The name of the sensor, e.g. 'Deflection-Sensor bottom-left'.
        measurand : str, optional
            Defines what the sensor measures, e.g. 'deflection'.
        unit : str, optional
            Defines what unit is associated with the sensor's measurements, for
            example 'mm'.
        x : float, int, numpy.ndarray, None, optional
            Positional x-coordinate of the sensor. When given, the coords-
            argument must be None.
        y : float, int, numpy.ndarray, None, optional
            Positional y-coordinate of the sensor. When given, the coords-
            argument must be None.
        z : float, int, numpy.ndarray, None, optional
            Positional z-coordinate of the sensor. When given, the coords-
            argument must be None.
        coords : numpy.ndarray, optional
            Some or all of the coordinates x, y, z concatenated as an array.
            Each row corresponds to a constant coordinate. Which row corresponds
            to which coordinate is defined via the order-argument. When given,
            the arguments x, y and z must be None.
        order : tuple[str], optional
            Only relevant when coords is given. Defines which row in coords
            corresponds to which coordinate. For example, order=('x', 'y', 'z')
            means that the 1st row are x-coordinates, the 2nd row are y-coords
            and the 3rd row are the z-coordinates.
        """

        # basic attributes
        self.name = name
        self.measurand = measurand
        self.unit = unit

        # translate the input to a coords-array
        self.coords, self._order = process_spatial_coordinates(
            x=x, y=y, z=z, coords=coords, order=order)

        # this contains the information which row contains which coordinate
        self.index_dict = {coord: i for i, coord in enumerate(self._order)}

    @property
    def x(self):
        """Provides x-coords as attribute without copying them from coords."""
        return self.coords[self.index_dict['x']] if 'x' in self._order else None

    @property
    def y(self):
        """Provides y-coords as attribute without copying them from coords."""
        return self.coords[self.index_dict['y']] if 'y' in self._order else None

    @property
    def z(self):
        """Provides z-coords as attribute without copying them from coords."""
        return self.coords[self.index_dict['z']] if 'z' in self._order else None

    @property
    def order(self):
        """Provides read-access to privat attribute self._order."""
        return self._order



# third party imports
import numpy as np

# local imports
from probeye.subroutines import process_spatial_coordinates


class SpatialExponentialCorrelationModel:
    """
    Represents a spatial correlation model with an exponential kernel. It
    contains the functionality to compute the covariance matrix over a static
    (i.e. constant for all experiments) grid of coordinates in 1D, 2D or 3D.
    """

    def __init__(self, x=None, y=None, z=None, coords=None,
                 order=('x', 'y', 'z')):
        """
        Parameter
        ---------
        x : float, int, numpy.ndarray, None, optional
            Positional x-coordinate. When given, coords must be None.
        y : float, int, numpy.ndarray, None, optional
            Positional y-coordinate. When given, coords must be None.
        z : float, int, numpy.ndarray, None, optional
            Positional z-coordinate. When given, coords must be None.
        coords : numpy.ndarray, optional
            Some or all of the coordinates x, y, z concatenated as an array.
            Each row corresponds to one coordinate. For example, row 1 might
            contain all x-coordinates. Which row corresponds to which coordinate
            is defined via the order-argument. When the coords-argument is
            given, all 3 arguments x, y and z must be None.
        order : tuple[str], optional
            Only relevant when coords is given. Defines which row in coords
            corresponds to which coordinate. For example, order=('x', 'y', 'z')
            means that the 1st row are x-coordinates, the 2nd row are y-coords
            and the 3rd row are the z-coordinates.
        """
        # translate the spatial input to a coords-array
        self.coords, self._order = process_spatial_coordinates(
            x=x, y=y, z=z, coords=coords, order=order)
        self.n_coords, self.n = self.coords.shape

        # on position (i, j) in self.distance array will be denoted the distance
        # between point i with coords[i, :] and point j with coords[j, :]
        distance_array = np.zeros((self.n, self.n))
        for i in range(self.n_coords):
            v = self.coords[i, :]
            v_in_columns = np.tile(v.reshape((self.n, -1)), self.n)
            v_in_rows = v_in_columns.transpose()  # each row is v
            distance_array += np.square(v_in_columns - v_in_rows)
        self.distance_array = np.sqrt(distance_array)

    @staticmethod
    def check_prms(prms):
        """
        Checks if the numeric values provided via prms are valid for computing
        the covariance matrix of this model.

        Parameters
        ----------
        prms : dict
            Contains the names of the correlation model's parameters as keys
            and the corresponding numeric values as values.

        Returns
        -------
        bool
            True, when all values are valid. False otherwise.
        """
        # if no correlation is defined in the noise model, l_corr will not be
        # provided in the input for the log-likelihood contribution
        if 'l_corr' in prms:
            if prms['l_corr'] <= 0:
                return False
        # sigma, however, must always be passed
        if prms['std'] <= 0:
            return False
        return True

    def __call__(self, prms):
        """
        Returns the covariance matrix based on the correlation model.

        Parameters
        ----------
        prms : dict
            Contains the names of the correlation model's parameters as keys
            and the corresponding numeric values as values.

        Returns
        -------
        numpy.ndarray
            The covariance matrix based on the given parameters. The shape of
            this array is (self.n, self.n).
        """
        corr = np.exp(-self.distance_array / prms['l_corr'])
        return prms['std'] ** 2 * corr
